"""Enhanced semantic validation for YARA AST."""

from __future__ import annotations

from typing import TYPE_CHECKING

from yaraast.types.module_loader import ModuleLoader
from yaraast.types.semantic_validator_core import ValidationError, ValidationResult
from yaraast.types.semantic_validator_functions import FunctionCallValidator
from yaraast.types.semantic_validator_helpers import populate_env_for_file, populate_env_for_rule
from yaraast.types.semantic_validator_strings import (
    StringIdentifierValidator,
    UndefinedStringDetector,
)
from yaraast.types.type_system import TypeChecker, TypeEnvironment

if TYPE_CHECKING:
    from yaraast.ast.base import YaraFile
    from yaraast.ast.expressions import Expression
    from yaraast.ast.rules import Rule


class SemanticValidator:
    """Comprehensive semantic validator for YARA AST."""

    def __init__(self) -> None:
        self.module_loader = ModuleLoader()

    def validate(self, ast: YaraFile) -> ValidationResult:
        """Perform complete semantic validation on YARA file."""
        result = ValidationResult()
        env = TypeEnvironment()

        populate_env_for_file(ast, env)

        string_validator = StringIdentifierValidator(result)
        for rule in ast.rules:
            string_validator.visit(rule)

        # Detect undefined strings referenced in conditions
        undefined_detector = UndefinedStringDetector(result)
        for rule in ast.rules:
            undefined_detector.check_rule(rule)

        function_validator = FunctionCallValidator(result, env)
        for rule in ast.rules:
            if rule.condition:
                function_validator.visit(rule.condition)

        type_checker = TypeChecker()
        type_errors = type_checker.check(ast)

        for error_msg in type_errors:
            result.add_error(
                error_msg,
                suggestion="Check variable types and function signatures",
            )

        return result

    def validate_rule(
        self,
        rule: Rule,
        env: TypeEnvironment | None = None,
    ) -> ValidationResult:
        """Validate a single rule."""
        result = ValidationResult()

        if env is None:
            env = TypeEnvironment()

        populate_env_for_rule(rule, env)

        string_validator = StringIdentifierValidator(result)
        string_validator.visit(rule)

        if rule.condition:
            function_validator = FunctionCallValidator(result, env)
            function_validator.visit(rule.condition)

        return result

    def validate_expression(
        self,
        expr: Expression,
        env: TypeEnvironment | None = None,
    ) -> ValidationResult:
        """Validate a single expression."""
        result = ValidationResult()

        if env is None:
            env = TypeEnvironment()

        function_validator = FunctionCallValidator(result, env)
        function_validator.visit(expr)

        return result


# Convenience functions for easy usage
def validate_yara_file(ast: YaraFile) -> ValidationResult:
    """Validate YARA file with comprehensive semantic checks."""
    validator = SemanticValidator()
    return validator.validate(ast)


def validate_yara_rule(
    rule: Rule,
    env: TypeEnvironment | None = None,
) -> ValidationResult:
    """Validate a single YARA rule."""
    validator = SemanticValidator()
    return validator.validate_rule(rule, env)


def check_string_uniqueness(rule: Rule) -> list[ValidationError]:
    """Check string identifier uniqueness within a rule."""
    result = ValidationResult()
    validator = StringIdentifierValidator(result)
    validator.visit(rule)
    return result.errors


def check_function_calls(
    expr: Expression,
    env: TypeEnvironment,
) -> list[ValidationError]:
    """Check function calls in an expression."""
    result = ValidationResult()
    validator = FunctionCallValidator(result, env)
    validator.visit(expr)
    return result.errors



"""Document symbols provider for YARA Language Server."""

from __future__ import annotations

import logging
import time

from lsprotocol.types import DocumentSymbol, Range

from yaraast.lsp.runtime import DocumentContext, LspRuntime
from yaraast.lsp.symbol_tree_builder import (
    build_document_symbols,
    find_closing_brace as _find_closing_brace_impl,
    find_line_containing as _find_line_containing_impl,
    make_range as _make_range_impl,
)

logger = logging.getLogger(__name__)


class SymbolsProvider:
    """Provides document symbols (outline view)."""

    def __init__(self, runtime: LspRuntime | None = None) -> None:
        self.runtime = runtime

    def get_symbols(self, text: str, uri: str | None = None) -> list[DocumentSymbol]:
        """
        Get document symbols for the given YARA file.

        Args:
            text: The YARA source code

        Returns:
            List of document symbols
        """
        started = time.perf_counter()

        try:
            if self.runtime and uri:
                doc = self.runtime.ensure_document(uri, text)
            else:
                doc = DocumentContext(uri or "", text)
            cached = doc.get_cached("lsp:document_symbols")
            if cached is not None:
                return cached
            ast = doc.ast()
            if ast is None:
                return []
            lines = text.split("\n")
            symbols = build_document_symbols(doc, lines)

        except Exception:
            logger.debug("Operation failed in %s", __name__, exc_info=True)
            # If parsing fails, return empty symbols
            symbols = []

        doc.set_cached("lsp:document_symbols", symbols)
        if self.runtime is not None:
            self.runtime.record_latency(
                "document_symbols", (time.perf_counter() - started) * 1000.0
            )
        return symbols

    def _find_line_containing(
        self,
        lines: list[str],
        text: str,
        start: int = 0,
    ) -> int:
        """Find the line number containing the given text."""
        return _find_line_containing_impl(lines, text, start)

    def _find_closing_brace(self, lines: list[str], start: int) -> int:
        """Find the closing brace for a rule."""
        return _find_closing_brace_impl(lines, start)

    def _make_range(self, start_line: int, start_char: int, end_line: int, end_char: int) -> Range:
        """Create an LSP Range."""
        return _make_range_impl(start_line, start_char, end_line, end_char)



"""Streaming parser for large YARA files."""

from __future__ import annotations

import mmap
from pathlib import Path
from typing import TYPE_CHECKING, Any

from yaraast.dialects import YaraDialect
from yaraast.parser.parser import Parser
from yaraast.performance.streaming_mmap import iter_rule_texts_from_mmap
from yaraast.performance.streaming_result_builders import (
    build_error_parse_result,
    build_file_parse_result,
    build_rule_parse_result,
    default_streaming_stats,
    timed_now,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    import io

    from yaraast.ast.rules import Rule


class StreamingParser:
    """Parse large YARA files efficiently using streaming."""

    def __init__(
        self,
        buffer_size: int = 8192,
        max_memory_mb: int | None = None,
        enable_gc: bool = False,
        progress_callback: Callable | None = None,
        dialect: YaraDialect | None = None,
        dialect_parser_factory: Callable | None = None,
    ) -> None:
        """Initialize streaming parser.

        Args:
            buffer_size: Size of read buffer in bytes
            max_memory_mb: Maximum memory usage in MB
            enable_gc: Enable garbage collection between files
            progress_callback: Progress callback function
            dialect: YARA dialect to use (auto-detects if None)
            dialect_parser_factory: Optional factory(text, dialect) -> YaraFile
                for parsing non-standard dialects without circular imports

        """
        self.buffer_size = buffer_size
        self.max_memory_mb = max_memory_mb
        self.enable_gc = enable_gc
        self.progress_callback = progress_callback
        self.dialect = dialect
        self._dialect_parser_factory = dialect_parser_factory
        self.parser = Parser()
        self._cancelled = False
        self._stats = default_streaming_stats()

    def parse_file(
        self,
        file_path: str | Path,
        callback: Callable[[Rule], None] | None = None,
    ) -> Iterator[Rule]:
        """Parse a YARA file in streaming fashion.

        Args:
            file_path: Path to YARA file
            callback: Optional callback for each parsed rule

        Yields:
            Parsed rules one at a time

        """
        file_path = Path(file_path)

        with (
            open(file_path, "rb") as f,
            mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as mmapped_file,
        ):
            # Use memory mapping for large files
            yield from self._parse_mmap(mmapped_file, callback)

    def parse_stream(
        self,
        stream: io.IOBase,
        callback: Callable[[Rule], None] | None = None,
    ) -> Iterator[Rule]:
        """Parse a stream of YARA content.

        Args:
            stream: Input stream
            callback: Optional callback for each parsed rule

        Yields:
            Parsed rules one at a time

        """
        buffer = ""
        rule_buffer = []
        in_rule = False
        brace_count = 0

        while True:
            chunk = stream.read(self.buffer_size)
            if not chunk:
                break

            if isinstance(chunk, bytes):
                chunk = chunk.decode("utf-8", errors="replace")

            buffer += chunk
            self._stats["bytes_processed"] += len(chunk)

            # Process lines
            lines = buffer.split("\n")
            buffer = lines[-1]  # Keep incomplete line

            for line in lines[:-1]:
                stripped = line.strip()

                # Track rule boundaries
                if stripped.startswith("rule ") and not in_rule:
                    in_rule = True
                    rule_buffer = [line]
                    brace_count = line.count("{") - line.count("}")
                elif in_rule:
                    rule_buffer.append(line)
                    brace_count += line.count("{") - line.count("}")

                    # Complete rule found
                    if brace_count == 0 and "}" in line:
                        rule_text = "\n".join(rule_buffer)
                        rule = self._parse_rule_text(rule_text)
                        if rule:
                            self._stats["rules_parsed"] += 1
                            if callback:
                                callback(rule)
                            yield rule

                        in_rule = False
                        rule_buffer = []

        # Handle any remaining buffer
        if buffer and in_rule:
            rule_buffer.append(buffer)
            rule_text = "\n".join(rule_buffer)
            rule = self._parse_rule_text(rule_text)
            if rule:
                self._stats["rules_parsed"] += 1
                if callback:
                    callback(rule)
                yield rule

    def parse_file_chunked(
        self,
        file_path: str | Path,
        chunk_size: int = 100,
    ) -> Iterator[list[Rule]]:
        """Parse file and yield rules in chunks.

        Args:
            file_path: Path to YARA file
            chunk_size: Number of rules per chunk

        Yields:
            Lists of parsed rules

        """
        chunk = []

        for rule in self.parse_file(file_path):
            chunk.append(rule)
            if len(chunk) >= chunk_size:
                yield chunk
                chunk = []

        # Yield remaining rules
        if chunk:
            yield chunk

    def parse_rules_from_file(self, file_path: Path) -> Iterator[Any]:
        """Parse individual rules from a file (yields one result per rule)."""
        try:
            start_time = timed_now()

            for rule in self.parse_file(file_path):
                if self._cancelled:
                    break

                parse_time = timed_now() - start_time
                yield build_rule_parse_result(file_path, rule, parse_time)
                start_time = timed_now()
        except Exception as e:
            yield build_error_parse_result(file_path, e)

    def parse_files(self, file_paths: list[Path]) -> Iterator[Any]:
        """Parse multiple files (yields one result per file)."""
        for idx, file_path in enumerate(file_paths, 1):
            if self._cancelled:
                break

            try:
                start_time = timed_now()
                content = Path(file_path).read_text()
                ast = self.parser.parse(content)
                parse_time = timed_now() - start_time

                self._stats["files_processed"] += 1
                self._stats["files_successful"] += 1
                self._stats["total_parse_time"] += parse_time
                self._stats["rules_parsed"] += len(ast.rules)

                # Call progress callback if provided
                if self.progress_callback:
                    self.progress_callback(idx, len(file_paths), str(file_path))

                yield build_file_parse_result(file_path, ast, parse_time)
            except Exception as e:
                self._stats["files_processed"] += 1
                self._stats["parse_errors"] += 1

                yield build_error_parse_result(file_path, e)
            finally:
                self._maybe_collect_garbage()

    def parse_directory(
        self,
        dir_path: Path,
        pattern: str = "*.yar",
        recursive: bool = False,
    ) -> Iterator[Any]:
        """Parse all files in a directory."""
        files = list(dir_path.rglob(pattern)) if recursive else list(dir_path.glob(pattern))

        yield from self.parse_files(files)

    def get_statistics(self) -> dict[str, Any]:
        """Get parser statistics."""
        return dict(self._stats)

    def cancel(self) -> None:
        """Cancel parsing."""
        self._cancelled = True

    def parse_with_progress(
        self,
        file_path: str | Path,
        progress_callback: Callable[[int, int], None],
    ) -> list[Rule]:
        """Parse file with progress reporting.

        Args:
            file_path: Path to YARA file
            progress_callback: Callback(bytes_processed, total_bytes)

        Returns:
            List of parsed rules

        """
        file_path = Path(file_path)
        file_size = file_path.stat().st_size
        rules = []

        def rule_callback(rule: Rule) -> None:
            rules.append(rule)
            progress_callback(self._stats["bytes_processed"], file_size)

        list(self.parse_file(file_path, rule_callback))
        return rules

    def _maybe_collect_garbage(self) -> None:
        """Run garbage collection based on configuration."""
        if self.enable_gc or self._memory_limit_exceeded():
            import gc

            gc.collect()

    def _memory_limit_exceeded(self) -> bool:
        """Check if current RSS exceeds the configured memory limit."""
        if self.max_memory_mb is None:
            return False
        try:
            import os

            import psutil

            process = psutil.Process(os.getpid())
            rss_mb = process.memory_info().rss / 1024 / 1024
            return rss_mb > self.max_memory_mb
        except Exception:
            return False

    def _parse_mmap(
        self,
        mmapped_file: mmap.mmap,
        callback: Callable[[Rule], None] | None = None,
    ) -> Iterator[Rule]:
        """Parse memory-mapped file content using proper tokenization.

        This method uses the Lexer to properly identify rule boundaries,
        avoiding issues with braces in strings, regexes, or comments.
        """
        for rule_text in iter_rule_texts_from_mmap(mmapped_file):
            rule = self._parse_rule_text(rule_text)
            if rule:
                self._stats["rules_parsed"] += 1
                if callback:
                    callback(rule)
                yield rule

    def _parse_rule_text(self, rule_text: str) -> Rule | None:
        """Parse a single rule text using the appropriate dialect parser."""
        try:
            dialect = self.dialect
            if dialect is not None and dialect != YaraDialect.YARA:
                if self._dialect_parser_factory is not None:
                    result = self._dialect_parser_factory(rule_text, dialect)
                    if hasattr(result, "rules") and result.rules:
                        return result.rules[0]
            else:
                # Use standard parser (default, fastest path)
                yara_file = self.parser.parse(rule_text)
                if yara_file.rules:
                    return yara_file.rules[0]
        except Exception:
            self._stats["parse_errors"] += 1

        return None

    def reset_statistics(self) -> None:
        """Reset parsing statistics."""
        self._stats = default_streaming_stats()

    def estimate_memory_usage(self, file_path: str | Path) -> dict[str, Any]:
        """Estimate memory usage for parsing a file.

        Args:
            file_path: Path to YARA file

        Returns:
            Memory usage estimates

        """
        file_path = Path(file_path)
        file_size = file_path.stat().st_size

        # Rough estimates based on experience
        estimated_ast_size = file_size * 3  # AST typically 3x file size
        estimated_peak = file_size * 5  # Peak during parsing

        return {
            "file_size_mb": file_size / 1024 / 1024,
            "estimated_ast_mb": estimated_ast_size / 1024 / 1024,
            "estimated_peak_mb": estimated_peak / 1024 / 1024,
            "streaming_buffer_mb": self.buffer_size / 1024 / 1024,
        }



"""Extra branch coverage for metrics/string_diagrams_render.py."""

from __future__ import annotations

from yaraast.ast.strings import (
    HexAlternative,
    HexByte,
    HexJump,
    HexString,
    HexWildcard,
    PlainString,
    RegexString,
)
from yaraast.metrics.string_diagrams_render import (
    StringDiagramRenderMixin,
    analyze_string_patterns,
    create_hex_diagram,
    create_regex_diagram,
    generate_pattern_report,
    generate_string_diagram,
)


class _Renderer(StringDiagramRenderMixin):
    pass


def test_render_mixin_branches_plain_hex_regex() -> None:
    r = _Renderer()

    plain = PlainString(identifier="$a", value="abc", modifiers=[])
    assert "Length: 3" in r._generate_plain_diagram(plain)

    plain_mod = PlainString(identifier="$am", value="abc", modifiers=["ascii"])
    assert "Modifiers: ascii" in r._generate_plain_diagram(plain_mod)

    hx = HexString(
        identifier="$h",
        tokens=[
            HexByte(value=0x41),
            HexWildcard(),
            HexJump(min_jump=2, max_jump=2),
            HexJump(min_jump=1, max_jump=3),
            HexAlternative(alternatives=[0x90, 0x91]),
        ],
        modifiers=["wide"],
    )
    hex_diag = r._generate_hex_diagram(hx)
    assert "41" in hex_diag
    assert "??" in hex_diag
    assert "[2]" in hex_diag
    assert "[1-3]" in hex_diag
    assert "(90|91)" in hex_diag
    assert "Modifiers: wide" in hex_diag

    rg = RegexString(identifier="$r", regex="^(ab)+$", modifiers=["nocase"])
    rg_diag = r._generate_regex_diagram(rg)
    assert "Capture Groups:" in rg_diag
    assert "Quantifiers:" in rg_diag
    assert "Modifiers: nocase" in rg_diag

    rg_simple = RegexString(identifier="$r2", regex="abc", modifiers=[])
    rg_simple_diag = r._generate_regex_diagram(rg_simple)
    assert "Capture Groups:" not in rg_simple_diag
    assert "Quantifiers:" not in rg_simple_diag


def test_render_convenience_functions_and_reports() -> None:
    plain = PlainString(identifier="$a", value="abc", modifiers=[])
    conv = generate_string_diagram(plain)
    assert "PlainString" in conv

    hex_line = create_hex_diagram(
        [
            HexByte(value=0xAA),
            HexWildcard(),
            HexJump(min_jump=2, max_jump=2),
            HexAlternative(alternatives=[0x01, 0x02]),
        ]
    )
    assert hex_line == "AA ?? [2] (01|02)"

    regex_diag = create_regex_diagram("^(ab)+[0-9]$")
    assert "Pattern:" in regex_diag
    assert "Capture groups:" in regex_diag
    assert "Quantifiers:" in regex_diag
    assert "Anchors:" in regex_diag
    assert "Character classes:" in regex_diag

    strings = [
        PlainString(identifier="$a", value="one", modifiers=["ascii"]),
        PlainString(identifier="$b", value="one", modifiers=[]),
        RegexString(identifier="$r", regex="ab+", modifiers=[]),
        HexString(identifier="$h", tokens=[HexByte(value=0x41)], modifiers=[]),
    ]

    analysis = analyze_string_patterns(strings)
    assert analysis["types"]["plain"] == 2
    assert analysis["types"]["regex"] == 1
    assert analysis["types"]["hex"] == 1
    assert "one" in analysis["patterns"]["duplicates"]

    report = generate_pattern_report(strings)
    assert report["summary"]["total"] == 4
    assert report["summary"]["unique_patterns"] == 3
    assert len(report["details"]) == 4



"""Recovery helpers for the error-tolerant parser."""

from __future__ import annotations

import re
from typing import Any

from yaraast.ast.base import Location
from yaraast.ast.expressions import BooleanLiteral, Identifier
from yaraast.ast.meta import Meta
from yaraast.ast.rules import Import, Include, Rule
from yaraast.ast.strings import PlainString


def parse_import_line(parser, line: str, line_num: int) -> Import | None:
    match = re.match(r'import\s+"([^"]+)"', line)
    if match:
        node = Import(match.group(1))
        node.location = Location(
            line=line_num + 1,
            column=match.start(1) + 1,
            end_line=line_num + 1,
            end_column=match.end(1) + 1,
        )
        return node
    match = re.match(r"import\s+(\w+)", line)
    if match:
        node = Import(match.group(1))
        node.location = Location(
            line=line_num + 1,
            column=match.start(1) + 1,
            end_line=line_num + 1,
            end_column=match.end(1) + 1,
        )
        return node
    parser._add_error(f"Invalid import statement: {line}", line_num, 0)
    return None


def parse_include_line(parser, line: str, line_num: int) -> Include | None:
    match = re.match(r'include\s+"([^"]+)"', line)
    if match:
        node = Include(match.group(1))
        node.location = Location(
            line=line_num + 1,
            column=match.start(1) + 1,
            end_line=line_num + 1,
            end_column=match.end(1) + 1,
        )
        return node
    parser._add_error(f"Invalid include statement: {line}", line_num, 0)
    return None


def create_rule_from_body(
    parser, name: str, tags: list[str], body_lines: list[str], start_line: int = 0
) -> Rule:
    rule = Rule(name=name, condition=BooleanLiteral(True))
    rule.tags = tags
    rule.meta = []
    rule.strings = []
    header_text = parser.lines[start_line] if 0 <= start_line < len(parser.lines) else ""
    name_column = header_text.find(name) + 1 if name in header_text else 1
    end_line = start_line + max(len(body_lines) - 1, 0)
    end_text = (
        parser.lines[end_line]
        if 0 <= end_line < len(parser.lines)
        else (body_lines[-1] if body_lines else header_text)
    )
    rule.location = Location(
        line=start_line + 1, column=name_column, end_line=end_line + 1, end_column=len(end_text) + 1
    )
    section = None
    for offset, body_line in enumerate(body_lines, start=1):
        section = parse_body_line(parser, rule, body_line, section, start_line + offset)
    return rule


def parse_body_line(
    parser, rule: Rule, body_line: str, current_section: str | None, line_num: int
) -> str | None:
    stripped = body_line.strip()
    if stripped.startswith("meta:"):
        return "meta"
    if stripped.startswith("strings:"):
        return "strings"
    if stripped.startswith("condition:"):
        condition_text = stripped[10:].strip()
        if condition_text:
            rule.condition = parse_condition(parser, condition_text, line_num, body_line)
        return "condition"
    if not stripped or stripped.startswith("}"):
        return current_section
    parse_section_content(parser, rule, stripped, current_section, line_num, body_line)
    return current_section


def parse_section_content(
    parser, rule: Rule, line: str, section: str | None, line_num: int, raw_line: str
) -> None:
    if section == "meta":
        meta_item = parse_meta_line(parser, line, line_num, raw_line)
        if meta_item:
            rule.meta.append(meta_item)
    elif section == "strings":
        string_def = parse_string_line(parser, line, line_num, raw_line)
        if string_def:
            rule.strings.append(string_def)
    elif section == "condition":
        rule.condition = parse_condition(parser, line, line_num, raw_line)


def parse_meta_line(
    parser, line: str, line_num: int | None = None, raw_line: str | None = None
) -> Meta | None:
    for pattern, converter in (
        (r'(\w+)\s*=\s*"([^"]*)"', lambda m: m.group(2)),
        (r"(\w+)\s*=\s*(\d+)", lambda m: int(m.group(2))),
        (r"(\w+)\s*=\s*(true|false)", lambda m: m.group(2).lower() == "true"),
    ):
        match = re.match(pattern, line, re.IGNORECASE)
        if match:
            node = Meta(match.group(1), converter(match))
            set_recovered_location(parser, node, line_num, raw_line, match.start(1), match.end(2))
            return node
    return None


def parse_string_line(
    parser, line: str, line_num: int | None = None, raw_line: str | None = None
) -> PlainString | None:
    for pattern in (
        r'(\$\w+)\s*=\s*"([^"]*)"',
        r"(\$\w+)\s*=\s*{([^}]+)}",
        r"(\$\w+)\s*=\s*/([^/]+)/",
    ):
        match = re.match(pattern, line)
        if match:
            node = PlainString(identifier=match.group(1), value=match.group(2))
            set_recovered_location(
                parser, node, line_num, raw_line, match.start(1), match.end(2) + 1
            )
            return node
    return None


def parse_condition(
    parser, condition_text: str, line_num: int | None = None, raw_line: str | None = None
) -> Any:
    condition_text = condition_text.strip()
    if condition_text == "true":
        node = BooleanLiteral(True)
    elif condition_text == "false":
        node = BooleanLiteral(False)
    else:
        node = Identifier(condition_text)
    start = raw_line.find(condition_text) if raw_line else 0
    if start < 0:
        start = 0
    set_recovered_location(parser, node, line_num, raw_line, start, start + len(condition_text))
    return node


def set_recovered_location(
    parser, node: Any, line_num: int | None, raw_line: str | None, start_col: int, end_col: int
) -> Any:
    if line_num is None:
        return node
    if 0 <= line_num < len(parser.lines):
        line_text = raw_line if raw_line is not None else parser.lines[line_num]
    else:
        line_text = raw_line or ""
    actual_start = max(0, start_col)
    actual_end = max(actual_start + 1, min(len(line_text), end_col))
    node.location = Location(
        line=line_num + 1, column=actual_start + 1, end_line=line_num + 1, end_column=actual_end + 1
    )
    return node



"""Reporting helpers for parse CLI command."""

from __future__ import annotations

from yaraast.cli.parse_output_services import _generate_output_by_format, _report_parsing_errors
from yaraast.cli.utils import print_cli_error


def report_parsing_errors(lexer_errors, parser_errors, ast) -> None:
    """Report parsing errors."""
    _report_parsing_errors(lexer_errors, parser_errors, ast)


def generate_output_by_format(ast, output_format: str, output: str | None) -> None:
    """Generate output for the requested format."""
    _generate_output_by_format(ast, output_format, output)


def display_cli_error(console, error: Exception) -> None:
    """Display CLI error using standard formatting."""
    print_cli_error(console, error)



"""Tests for simple roundtrip serialization utilities."""

from __future__ import annotations

from pathlib import Path

from yaraast.ast.expressions import (
    BinaryExpression,
    BooleanLiteral,
    DoubleLiteral,
    Identifier,
    IntegerLiteral,
    StringIdentifier,
    StringLiteral,
    UnaryExpression,
)
from yaraast.parser import Parser
from yaraast.serialization.simple_roundtrip import (
    SimpleRoundTrip,
    SimpleRoundtripSerializer,
    simple_roundtrip_test,
)


def _sample_yara_rule() -> str:
    return """
import "pe"

rule test_roundtrip : alpha beta {
    meta:
        author = "unit"
        active = true
        count = 3
    strings:
        $a = "abc"
        $b = { 6A 40 68 00 30 00 00 }
        $c = /foo.*/
    condition:
        any of them
}
"""


def test_simple_roundtrip_serializer_in_memory() -> None:
    parser = Parser()
    ast = parser.parse(_sample_yara_rule())

    serializer = SimpleRoundtripSerializer()
    data = serializer.serialize(ast)
    restored = serializer.deserialize(data)

    assert data["type"] == "YaraFile"
    assert restored is not None

    ok, diff = serializer.validate_roundtrip(ast)
    assert isinstance(ok, bool)
    assert "original_code" in diff


def test_simple_roundtrip_serializer_file_io(tmp_path: Path) -> None:
    parser = Parser()
    ast = parser.parse(_sample_yara_rule())

    serializer = SimpleRoundtripSerializer()
    out_path = tmp_path / "roundtrip.json"

    serializer.serialize_to_file(ast, out_path)
    loaded = serializer.deserialize_from_file(out_path)

    assert loaded is not None


def test_simple_roundtrip_helpers(tmp_path: Path) -> None:
    helper = SimpleRoundTrip()

    yara_code = _sample_yara_rule()
    ok, orig, regen = helper.test(yara_code)
    assert isinstance(ok, bool)
    assert orig is not None or regen is None

    batch = helper.test_batch([yara_code, yara_code])
    assert len(batch) == 2

    file_path = tmp_path / "sample.yar"
    file_path.write_text(yara_code)
    ok2, _, _ = helper.test_file(file_path)
    assert isinstance(ok2, bool)

    results = helper.test_directory(tmp_path)
    assert results

    stats = helper.get_statistics()
    assert stats["total_tests"] >= 3

    report = simple_roundtrip_test(yara_code)
    assert report["round_trip_successful"] in {True, False}


def test_simple_roundtrip_serialize_primitives() -> None:
    serializer = SimpleRoundtripSerializer()

    primitives = [
        BooleanLiteral(True),
        IntegerLiteral(1),
        DoubleLiteral(3.14),
        StringLiteral("hi"),
        Identifier("foo"),
        StringIdentifier("$a"),
        UnaryExpression("not", BooleanLiteral(False)),
        BinaryExpression(IntegerLiteral(1), "+", IntegerLiteral(2)),
    ]

    for node in primitives:
        data = serializer.serialize(node)
        restored = serializer.deserialize(data)
        assert restored is not None

    # Unknown type fallback should return an Identifier
    restored = serializer.deserialize({"type": "UnknownNode", "data": "fallback"})
    assert isinstance(restored, Identifier)


def test_simple_roundtrip_error_path() -> None:
    report = simple_roundtrip_test("rule broken { condition: }")
    assert report["round_trip_successful"] is False



"""Options section validation mixin for YARA-L."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from yaraast.yaral.ast_nodes import OptionsSection


class OptionsValidationMixin:
    """Validate options section."""

    def _validate_options_section(self, node: OptionsSection) -> None:
        """Validate options section."""
        valid_options = [
            "allow_zero_values",
            "case_sensitive",
            "max_events",
            "max_matches",
            "timeout",
            "output_format",
        ]

        if hasattr(node, "options"):
            for key in node.options:
                if key not in valid_options:
                    self._add_warning(
                        "options",
                        f"Unknown option: {key}",
                        f"Valid options: {', '.join(valid_options)}",
                    )

    def visit_yaral_options_section(self, node: OptionsSection) -> None:
        self._validate_options_section(node)



"""Default visitor implementations for metrics modules."""

from __future__ import annotations

from typing import Any

from yaraast.visitor.defaults import DefaultASTVisitor


class MetricsVisitorBase(DefaultASTVisitor[Any]):
    """ASTVisitor with default no-op implementations."""

    def __init__(self, default: Any = None) -> None:
        super().__init__(default=default)



"""Additional coverage for semantic_validator module convenience paths."""

from __future__ import annotations

from yaraast.ast.base import YaraFile
from yaraast.ast.expressions import FunctionCall, Identifier
from yaraast.ast.rules import Rule
from yaraast.ast.strings import PlainString
from yaraast.types.semantic_validator import (
    SemanticValidator,
    check_function_calls,
    check_string_uniqueness,
    validate_yara_file,
    validate_yara_rule,
)
from yaraast.types.type_system import TypeEnvironment


def _rule(name: str = "r", with_condition: bool = True) -> Rule:
    return Rule(
        name=name,
        strings=[PlainString(identifier="$a", value="x"), PlainString(identifier="$a", value="y")],
        condition=Identifier("true") if with_condition else None,
    )


def test_validate_rule_with_and_without_env_and_condition() -> None:
    validator = SemanticValidator()

    r1 = _rule(with_condition=True)
    res1 = validator.validate_rule(r1)
    assert res1.errors  # duplicate string id

    env = TypeEnvironment()
    env.add_module("pe")
    r2 = _rule(with_condition=False)
    res2 = validator.validate_rule(r2, env)
    assert res2.errors  # duplicate id still caught


def test_validate_expression_and_convenience_functions() -> None:
    validator = SemanticValidator()
    expr = FunctionCall(function="pe.imphash", arguments=[])

    # env None branch
    res_expr = validator.validate_expression(expr)
    assert res_expr.errors

    rule = _rule()
    yf = YaraFile(rules=[rule])

    full = validate_yara_file(yf)
    assert full.errors

    single = validate_yara_rule(rule)
    assert single.errors

    uniq_errors = check_string_uniqueness(rule)
    assert uniq_errors

    fn_errors = check_function_calls(expr, TypeEnvironment())
    assert fn_errors



from __future__ import annotations

from pathlib import Path

from lsprotocol.types import Position

from yaraast.lsp.definition import DefinitionProvider
from yaraast.lsp.hover import HoverProvider
from yaraast.lsp.references import ReferencesProvider
from yaraast.lsp.rename import RenameProvider
from yaraast.lsp.runtime import LspRuntime, path_to_uri


def _pos(line: int, char: int) -> Position:
    return Position(line=line, character=char)


def test_hover_provider_supports_yaral_rule_metadata_and_sections() -> None:
    text = """
rule detect_login {
    meta:
        author = "sec"
        severity = "high"
    events:
        $e.metadata.event_type = "USER_LOGIN"
    match:
        $e over 5m
    condition:
        #e > 5
    outcome:
        $risk_score = 80
}

rule wrapper {
    condition:
        detect_login
}
""".lstrip()
    runtime = LspRuntime()
    uri = "file:///yaral_rule.yar"
    runtime.open_document(uri, text)
    provider = HoverProvider(runtime)

    hover = provider.get_hover(text, _pos(16, 10), uri)
    assert hover is not None
    value = hover.contents.value
    assert "**detect_login**" in value
    assert "Metadata" in value
    assert "**YARA-L:** events section present" in value
    assert "**YARA-L:** match section present" in value
    assert "**YARA-L:** outcome section present" in value


def test_definition_references_and_rename_work_for_yaral_rules_cross_file(tmp_path: Path) -> None:
    common = tmp_path / "common.yar"
    user = tmp_path / "user.yar"
    common.write_text(
        """
rule detect_login {
    events:
        $e.metadata.event_type = "USER_LOGIN"
    condition:
        #e > 5
}
""".lstrip(),
        encoding="utf-8",
    )
    user.write_text(
        """
rule wrapper {
    condition:
        detect_login
}
""".lstrip(),
        encoding="utf-8",
    )

    runtime = LspRuntime()
    runtime.update_config({"YARA": {"dialectMode": "yaral"}})
    runtime.set_workspace_folders([str(tmp_path)])

    uri = path_to_uri(user)
    text = user.read_text(encoding="utf-8")

    definition = DefinitionProvider(runtime).get_definition(text, _pos(2, 10), uri)
    assert definition is not None
    assert definition.uri == path_to_uri(common)

    records = ReferencesProvider(runtime).get_reference_records(text, _pos(2, 10), uri)
    assert {record.role for record in records} == {"declaration", "use"}
    assert {record.location.uri for record in records} == {path_to_uri(common), uri}

    edit = RenameProvider(runtime).rename(text, _pos(2, 10), "detect_login_new", uri)
    assert edit is not None and edit.changes is not None
    assert set(edit.changes) == {path_to_uri(common), uri}


def test_definition_and_hover_work_for_yarax_rule() -> None:
    text = """
rule helper { condition: true }
rule sample {
  condition:
    with $x = 1:
      helper
}
""".lstrip()
    runtime = LspRuntime()
    runtime.update_config({"YARA": {"dialectMode": "yarax"}})
    uri = "file:///yarax_rule.yar"
    runtime.open_document(uri, text)

    definition = DefinitionProvider(runtime).get_definition(text, _pos(4, 8), uri)
    assert definition is not None
    assert definition.range.start.line == 0

    hover = HoverProvider(runtime).get_hover(text, _pos(4, 8), uri)
    assert hover is not None
    assert "**helper**" in hover.contents.value



"""Token stream helpers for parser mixins."""

from __future__ import annotations

from yaraast.ast.base import ASTNode, Location
from yaraast.interfaces import IToken
from yaraast.lexer import TokenType


class TokenStreamMixin:
    """Mixin providing token stream helpers."""

    tokens: list[IToken]
    current: int

    def _match(self, *types: TokenType) -> bool:
        """Check if current token matches any of the given types."""
        for token_type in types:
            if self._check(token_type):
                self._advance()
                return True
        return False

    def _check(self, token_type: TokenType) -> bool:
        """Check if current token is of given type."""
        if self._is_at_end():
            return False
        return self._peek().type == token_type

    def _check_any(self, *types: TokenType) -> bool:
        """Check if current token matches any of the given types."""
        return any(self._check(t) for t in types)

    def _advance(self) -> IToken:
        """Consume current token and return it."""
        if not self._is_at_end():
            self.current += 1
        return self._previous()

    def _is_at_end(self) -> bool:
        """Check if we're at end of tokens."""
        return self._peek().type == TokenType.EOF

    def _peek(self) -> IToken:
        """Return current token without advancing."""
        return self.tokens[self.current]

    def _previous(self) -> IToken:
        """Return previous token."""
        if self.current <= 0:
            return self.tokens[0]
        return self.tokens[self.current - 1]

    def _token_span(self, token: IToken) -> int:
        length = getattr(token, "length", 0) or 0
        return length if length > 0 else max(1, len(str(token.value)))

    def _location_from_token(self, token: IToken) -> Location:
        return Location(
            line=token.line,
            column=token.column,
            end_line=token.line,
            end_column=token.column + self._token_span(token),
        )

    def _location_from_tokens(self, start_token: IToken, end_token: IToken) -> Location:
        return Location(
            line=start_token.line,
            column=start_token.column,
            end_line=end_token.line,
            end_column=end_token.column + self._token_span(end_token),
        )

    def _location_from_nodes(self, start_node: ASTNode, end_node: ASTNode) -> Location | None:
        start = getattr(start_node, "location", None)
        end = getattr(end_node, "location", None)
        if start is None or end is None:
            return None
        return Location(
            line=start.line,
            column=start.column,
            file=start.file or end.file,
            end_line=end.end_line or end.line,
            end_column=end.end_column or (end.column + 1),
        )

    def _set_node_location_from_token(self, node: ASTNode, token: IToken) -> ASTNode:
        node.location = self._location_from_token(token)
        return node

    def _set_node_location_from_tokens(
        self, node: ASTNode, start_token: IToken, end_token: IToken
    ) -> ASTNode:
        node.location = self._location_from_tokens(start_token, end_token)
        return node

    def _set_node_location_from_nodes(
        self, node: ASTNode, start_node: ASTNode, end_node: ASTNode
    ) -> ASTNode:
        location = self._location_from_nodes(start_node, end_node)
        if location is not None:
            node.location = location
        return node

    def _synthetic_token_from_location(self, location: Location):
        class _SyntheticToken:
            def __init__(self, loc: Location) -> None:
                self.line = loc.line
                self.column = loc.column
                if loc.end_line == loc.line and loc.end_column is not None:
                    self.length = max(1, loc.end_column - loc.column)
                else:
                    self.length = 1
                self.value = ""

        return _SyntheticToken(location)



"""Leaf rendering helpers for CodeGenerator visitor methods."""

from __future__ import annotations

from yaraast.codegen.generator_formatting import (
    escape_string_literal,
    format_boolean_literal,
    format_hex_jump,
    format_regex_literal,
)
from yaraast.codegen.generator_helpers import format_integer_literal


def visit_hex_byte(node) -> str:
    if isinstance(node.value, str):
        return node.value.upper()
    return f"{node.value:02X}"


def visit_hex_wildcard(node) -> str:
    return "??"


def visit_hex_jump(node) -> str:
    return format_hex_jump(node.min_jump, node.max_jump)


def visit_hex_alternative(generator, node) -> str:
    alts = []
    for alt in node.alternatives:
        alt_str = " ".join(generator.visit(token) for token in alt)
        alts.append(alt_str)
    return f"( {' | '.join(alts)} )"


def visit_hex_nibble(node) -> str:
    value_str = node.value.upper() if isinstance(node.value, str) else f"{node.value:X}"
    return f"{value_str}?" if node.high else f"?{value_str}"


def visit_identifier(node) -> str:
    return node.name


def visit_string_identifier(node) -> str:
    return node.name


def visit_string_wildcard(node) -> str:
    return node.pattern


def visit_string_count(node) -> str:
    return f"#{node.string_id}"


def visit_string_offset(generator, node) -> str:
    if node.index:
        return f"@{node.string_id}[{generator.visit(node.index)}]"
    return f"@{node.string_id}"


def visit_string_length(generator, node) -> str:
    if node.index:
        return f"!{node.string_id}[{generator.visit(node.index)}]"
    return f"!{node.string_id}"


def visit_integer_literal(node) -> str:
    return format_integer_literal(node.value)


def visit_double_literal(node) -> str:
    return str(node.value)


def visit_string_literal(node) -> str:
    return f'"{escape_string_literal(node.value)}"'


def visit_regex_literal(node) -> str:
    return format_regex_literal(node.pattern, node.modifiers)


def visit_boolean_literal(node) -> str:
    return format_boolean_literal(node.value)


def visit_meta(node) -> str:
    if isinstance(node.value, str):
        return f'{node.key} = "{node.value}"'
    if isinstance(node.value, bool):
        return f"{node.key} = {'true' if node.value else 'false'}"
    return f"{node.key} = {node.value}"


def visit_module_reference(node) -> str:
    return node.module


def visit_dictionary_access(generator, node) -> str:
    obj = generator.visit(node.object)
    if isinstance(node.key, str):
        return f'{obj}["{node.key}"]'
    return f"{obj}[{generator.visit(node.key)}]"


def visit_defined_expression(generator, node) -> str:
    return f"defined {generator.visit(node.expression)}"


def visit_string_operator_expression(generator, node) -> str:
    return f"{generator.visit(node.left)} {node.operator} {generator.visit(node.right)}"


def visit_comment(node) -> str:
    return f"// {node.text}"


def visit_comment_group(node) -> str:
    return "\n".join(f"// {c.text}" for c in node.comments)


def visit_extern_import(node) -> str:
    return f'import "{node.module_path}"'


def visit_extern_namespace(node) -> str:
    return f"namespace {node.name}"


def visit_extern_rule(node) -> str:
    modifiers = " ".join(str(m) for m in node.modifiers) if hasattr(node, "modifiers") else ""
    if modifiers:
        return f"{modifiers} rule {node.name}"
    return f"rule {node.name}"


def visit_extern_rule_reference(node) -> str:
    return node.name


def visit_in_rule_pragma(node) -> str:
    args_str = " " + " ".join(node.pragma.arguments) if node.pragma.arguments else ""
    return f"#pragma {node.pragma.name}{args_str}"


def visit_pragma(node) -> str:
    args_str = " " + " ".join(node.arguments) if node.arguments else ""
    return f"#pragma {node.name}{args_str}"


def visit_pragma_block(generator, node) -> str:
    return "\n".join(generator.visit(pragma) for pragma in node.pragmas)



"""Symbol resolution helpers for LSP document contexts."""

from __future__ import annotations

from typing import TYPE_CHECKING

from lsprotocol.types import Position

from yaraast.lsp.document_query_resolution_ast import resolve_symbol_from_ast
from yaraast.lsp.document_query_resolution_symbol_records import (
    prefer_symbol_resolution,
    resolve_symbol_from_symbol_records,
)
from yaraast.lsp.document_query_resolution_text import resolve_symbol_from_text_fallback
from yaraast.lsp.document_types import ResolvedSymbol

if TYPE_CHECKING:
    from yaraast.lsp.document_context import DocumentContext


def resolve_symbol(ctx: DocumentContext, position: Position) -> ResolvedSymbol | None:
    cache_key = f"resolve_symbol:{position.line}:{position.character}"
    cached = ctx.get_cached(cache_key)
    if cached is not None:
        return cached
    ast_resolved = resolve_symbol_from_ast(ctx, position)
    symbol_resolved = resolve_symbol_from_symbol_records(ctx, position)
    if symbol_resolved is not None and prefer_symbol_resolution(symbol_resolved, ast_resolved):
        ctx.set_cached(cache_key, symbol_resolved)
        return symbol_resolved
    if ast_resolved is not None:
        ctx.set_cached(cache_key, ast_resolved)
        return ast_resolved
    result = resolve_symbol_from_text_fallback(
        ctx,
        position,
        allow_generic_identifier=True,
    )
    if result is None:
        return None
    ctx.set_cached(cache_key, result)
    return result



"""YARA-L semantic validator."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from yaraast.yaral.validator_conditions import ConditionValidationMixin
from yaraast.yaral.validator_events import EventValidationMixin
from yaraast.yaral.validator_match import MatchValidationMixin
from yaraast.yaral.validator_meta import MetaValidationMixin
from yaraast.yaral.validator_options import OptionsValidationMixin
from yaraast.yaral.validator_outcomes import OutcomeValidationMixin
from yaraast.yaral.validator_rules import RuleValidationMixin
from yaraast.yaral.visitor_base import YaraLVisitor

if TYPE_CHECKING:
    from yaraast.yaral.ast_nodes import YaraLFile


@dataclass
class ValidationError:
    """Represents a validation error in YARA-L rules."""

    severity: str  # "error", "warning", "info"
    rule_name: str
    section: str
    message: str
    suggestion: str = ""

    def __str__(self) -> str:
        return f"[{self.severity.upper()}] {self.rule_name}/{self.section}: {self.message}"


class YaraLValidator(
    RuleValidationMixin,
    EventValidationMixin,
    MatchValidationMixin,
    ConditionValidationMixin,
    OutcomeValidationMixin,
    OptionsValidationMixin,
    MetaValidationMixin,
    YaraLVisitor[None],
):
    """Semantic validator for YARA-L rules."""

    # Valid UDM fields for different event types
    VALID_UDM_FIELDS = {
        "metadata": [
            "event_type",
            "event_timestamp",
            "product_name",
            "vendor_name",
            "description",
            "product_version",
            "product_event_type",
            "ingested_timestamp",
        ],
        "principal": [
            "hostname",
            "ip",
            "port",
            "mac",
            "user",
            "process",
            "file",
            "user_id",
            "email",
            "domain",
            "asset_id",
            "location",
        ],
        "target": [
            "hostname",
            "ip",
            "port",
            "mac",
            "user",
            "process",
            "file",
            "user_id",
            "email",
            "domain",
            "asset_id",
            "url",
            "resource",
        ],
        "observer": ["hostname", "ip", "port", "mac", "asset_id", "location"],
        "src": ["ip", "port", "mac", "hostname", "location", "country", "region"],
        "dst": ["ip", "port", "mac", "hostname", "location", "country", "region"],
        "network": [
            "application_protocol",
            "direction",
            "ip_protocol",
            "received_bytes",
            "sent_bytes",
            "session_duration",
            "session_id",
        ],
        "security_result": [
            "action",
            "category",
            "confidence",
            "severity",
            "rule_name",
            "detection_fields",
            "summary",
            "action_details",
        ],
        "about": ["file", "process", "user", "resource", "labels"],
    }

    # Valid aggregation functions
    VALID_AGGREGATIONS = [
        "count",
        "count_distinct",
        "sum",
        "avg",
        "min",
        "max",
        "array",
        "array_distinct",
        "string_concat",
    ]

    # Valid time units for windows
    VALID_TIME_UNITS = ["s", "m", "h", "d", "seconds", "minutes", "hours", "days"]

    def __init__(self) -> None:
        """Initialize validator."""
        self.errors: list[ValidationError] = []
        self.warnings: list[ValidationError] = []
        self.current_rule: str | None = None
        self.defined_events: set[str] = set()
        self.used_events: set[str] = set()
        self.defined_match_vars: set[str] = set()
        self.used_match_vars: set[str] = set()
        self.defined_outcome_vars: set[str] = set()

    def validate(self, ast: YaraLFile) -> tuple[list[ValidationError], list[ValidationError]]:
        """Validate YARA-L file and return errors and warnings.

        Args:
            ast: YARA-L AST to validate

        Returns:
            Tuple of (errors, warnings)
        """
        self.errors.clear()
        self.warnings.clear()
        self.visit(ast)
        return self.errors, self.warnings

    def _add_error(self, section: str, message: str, suggestion: str = "") -> None:
        """Add validation error."""
        self.errors.append(
            ValidationError(
                severity="error",
                rule_name=self.current_rule or "unknown",
                section=section,
                message=message,
                suggestion=suggestion,
            )
        )

    def _add_warning(self, section: str, message: str, suggestion: str = "") -> None:
        """Add validation warning."""
        self.warnings.append(
            ValidationError(
                severity="warning",
                rule_name=self.current_rule or "unknown",
                section=section,
                message=message,
                suggestion=suggestion,
            )
        )



"""Reporting helpers for AST diff CLI output."""

from __future__ import annotations

from pathlib import Path

from rich.console import Console

console = Console()


def display_no_changes(file1_path: Path, file2_path: Path) -> None:
    console.print(
        f"[green]No differences found between {file1_path.name} and {file2_path.name}[/green]",
    )


def display_diff_header(file1_path: Path, file2_path: Path) -> None:
    console.print(f"[blue]AST Diff: {file1_path.name} -> {file2_path.name}[/blue]")
    console.print("=" * 60)


def show_diff_summary(result) -> None:
    console.print("[yellow]Change Summary:[/yellow]")
    for change_type, count in result.change_summary.items():
        if count > 0:
            console.print(f"  - {change_type.replace('_', ' ').title()}: {count}")


def show_rule_changes(result) -> None:
    if result.added_rules:
        console.print(f"\n[green]+ Added Rules ({len(result.added_rules)}):[/green]")
        for rule in result.added_rules:
            console.print(f"  + {rule}")

    if result.removed_rules:
        console.print(f"\n[red]- Removed Rules ({len(result.removed_rules)}):[/red]")
        for rule in result.removed_rules:
            console.print(f"  - {rule}")

    if result.modified_rules:
        console.print(
            f"\n[yellow]Modified Rules ({len(result.modified_rules)}):[/yellow]",
        )
        for rule in result.modified_rules:
            console.print(f"  ~ {rule}")


def show_change_details(result, logical_only: bool, no_style: bool) -> None:
    if result.logical_changes:
        console.print(
            f"\n[red]Logical Changes ({len(result.logical_changes)}):[/red]",
        )
        for change in result.logical_changes:
            console.print(f"  - {change}")

    if result.structural_changes:
        console.print(
            f"\n[blue]Structural Changes ({len(result.structural_changes)}):[/blue]",
        )
        for change in result.structural_changes:
            console.print(f"  - {change}")

    if not logical_only and not no_style and result.style_only_changes:
        show_style_changes(result.style_only_changes)


def show_style_changes(style_changes: list) -> None:
    console.print(
        f"\n[dim]Style-Only Changes ({len(style_changes)}):[/dim]",
    )
    for change in style_changes[:10]:
        console.print(f"[dim]  - {change}[/dim]")
    if len(style_changes) > 10:
        console.print(
            f"[dim]  - ... and {len(style_changes) - 10} more style changes[/dim]",
        )


def show_change_significance(result) -> None:
    total_logical = len(result.added_rules) + len(result.removed_rules) + len(result.modified_rules)
    total_style = len(result.style_only_changes)

    if total_logical > 0:
        console.print(
            f"\n[yellow]This diff contains {total_logical} logical changes that affect rule behavior[/yellow]",
        )
    elif total_style > 0:
        console.print(
            f"\n[green]This diff contains only {total_style} style changes (no logic changes)[/green]",
        )



"""Models for round-trip serialization."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any


@dataclass
class FormattingInfo:
    """Information about original formatting to preserve."""

    indent_size: int = 4
    indent_style: str = "spaces"  # "spaces" or "tabs"
    line_endings: str = "\n"  # "\n", "\r\n", or "\r"
    blank_lines_before_rule: int = 1
    blank_lines_after_imports: int = 1
    blank_lines_after_includes: int = 1
    comment_style: str = "line"  # "line" (//) or "block" (/* */)
    preserve_spacing: bool = True
    preserve_alignment: bool = True

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary for serialization."""
        return {
            "indent_size": self.indent_size,
            "indent_style": self.indent_style,
            "line_endings": self.line_endings,
            "blank_lines_before_rule": self.blank_lines_before_rule,
            "blank_lines_after_imports": self.blank_lines_after_imports,
            "blank_lines_after_includes": self.blank_lines_after_includes,
            "comment_style": self.comment_style,
            "preserve_spacing": self.preserve_spacing,
            "preserve_alignment": self.preserve_alignment,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> FormattingInfo:
        """Create from dictionary."""
        return cls(**data)


@dataclass
class RoundTripMetadata:
    """Metadata for round-trip serialization."""

    original_source: str | None = None
    source_file: str | None = None
    parsed_at: str | None = None
    serializer_version: str = "1.0.0"
    formatting: FormattingInfo = field(default_factory=FormattingInfo)
    comments_preserved: bool = True
    formatting_preserved: bool = True
    parser_version: str | None = None

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary for serialization."""
        return {
            "original_source": self.original_source,
            "source_file": self.source_file,
            "parsed_at": self.parsed_at,
            "serializer_version": self.serializer_version,
            "formatting": self.formatting.to_dict(),
            "comments_preserved": self.comments_preserved,
            "formatting_preserved": self.formatting_preserved,
            "parser_version": self.parser_version,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> RoundTripMetadata:
        """Create from dictionary."""
        formatting_data = data.get("formatting", {})
        formatting = FormattingInfo.from_dict(formatting_data)

        return cls(
            original_source=data.get("original_source"),
            source_file=data.get("source_file"),
            parsed_at=data.get("parsed_at"),
            serializer_version=data.get("serializer_version", "1.0.0"),
            formatting=formatting,
            comments_preserved=data.get("comments_preserved", True),
            formatting_preserved=data.get("formatting_preserved", True),
            parser_version=data.get("parser_version"),
        )



"""Real integration tests for yarax/yaral/analyze CLI commands."""

from __future__ import annotations

from pathlib import Path
from textwrap import dedent

from click.testing import CliRunner

from yaraast.cli.commands.analyze import analyze
from yaraast.cli.commands.yaral import yaral
from yaraast.cli.commands.yarax import yarax


def _write(tmp_path: Path, name: str, content: str) -> str:
    p = tmp_path / name
    p.write_text(dedent(content), encoding="utf-8")
    return str(p)


def _sample_yara() -> str:
    return """
    rule sample {
        strings:
            $a = "abc"
        condition:
            $a
    }
    """


def _sample_yarax() -> str:
    return """
    rule sample_yarax {
        strings:
            $a = "abc"
        condition:
            with $x = #a:
                $a and $x > 0
    }
    """


def _sample_yaral() -> str:
    return """
    rule login_attempts {
        meta:
            author = "unit"
        events:
            $e.metadata.event_type = "USER_LOGIN"
        condition:
            $e
    }
    """


def _invalid_rule() -> str:
    return "rule broken { condition: "


def test_yarax_real_paths(tmp_path: Path) -> None:
    runner = CliRunner()

    yarax_file = _write(tmp_path, "ok.yarax", _sample_yarax())
    yara_file = _write(tmp_path, "ok.yar", _sample_yara())
    bad_file = _write(tmp_path, "bad.yarax", _invalid_rule())

    out_ast = tmp_path / "ast.txt"
    parse_ok = runner.invoke(yarax, ["parse", yarax_file, "-o", str(out_ast), "--show-features"])
    assert parse_ok.exit_code == 0
    assert out_ast.exists()
    assert "YARA-X Features Used" in parse_ok.output

    parse_bad = runner.invoke(yarax, ["parse", bad_file])
    assert parse_bad.exit_code != 0

    check_ok = runner.invoke(yarax, ["check", yara_file])
    assert check_ok.exit_code == 0

    check_bad = runner.invoke(yarax, ["check", bad_file, "--strict"])
    assert check_bad.exit_code != 0

    out_yarax = tmp_path / "conv.yarax"
    conv_to_yarax = runner.invoke(
        yarax, ["convert", yara_file, "--target", "yarax", "-o", str(out_yarax)]
    )
    assert conv_to_yarax.exit_code == 0
    assert out_yarax.exists()

    conv_to_yarax_stdout = runner.invoke(yarax, ["convert", yara_file, "--target", "yarax"])
    assert conv_to_yarax_stdout.exit_code == 0
    assert "rule" in conv_to_yarax_stdout.output.lower()

    conv_bad = runner.invoke(yarax, ["convert", bad_file, "--target", "yara"])
    assert conv_bad.exit_code != 0

    playground_file = runner.invoke(yarax, ["playground", "--file", yarax_file])
    assert playground_file.exit_code == 0
    assert "Successfully parsed" in playground_file.output

    playground_default = runner.invoke(yarax, ["playground"])
    assert playground_default.exit_code == 0
    assert "Example YARA-X code" in playground_default.output

    playground_bad = runner.invoke(yarax, ["playground", _invalid_rule()])
    assert playground_bad.exit_code == 0
    assert "Parse error" in playground_bad.output


def test_yaral_real_paths(tmp_path: Path) -> None:
    runner = CliRunner()

    good = _write(tmp_path, "ok.yaral", _sample_yaral())
    other = _write(
        tmp_path, "other.yaral", _sample_yaral().replace("login_attempts", "login_attempts_2")
    )
    bad = _write(tmp_path, "bad.yaral", _invalid_rule())
    bad_path = str(tmp_path)  # directory path triggers read error in command handlers

    parse_yaml = runner.invoke(yaral, ["parse", good, "--format", "yaml"])
    assert parse_yaml.exit_code == 0

    parse_text = runner.invoke(yaral, ["parse", good, "--format", "text"])
    assert parse_text.exit_code == 0

    parse_bad = runner.invoke(yaral, ["parse", bad])
    assert parse_bad.exit_code != 0

    validate_ok = runner.invoke(yaral, ["validate", good, "--json", "--strict"])
    assert validate_ok.exit_code == 0

    validate_bad = runner.invoke(yaral, ["validate", bad])
    assert validate_bad.exit_code != 0

    optimize_dry = runner.invoke(yaral, ["optimize", good, "--dry-run"])
    assert optimize_dry.exit_code == 0

    out_opt = tmp_path / "opt.yaral"
    optimize_stats = runner.invoke(yaral, ["optimize", good, "--stats", "-o", str(out_opt)])
    assert optimize_stats.exit_code == 0
    assert out_opt.exists()

    optimize_bad = runner.invoke(yaral, ["optimize", bad])
    assert optimize_bad.exit_code != 0

    generate_ok = runner.invoke(yaral, ["generate", good, "--format"])
    assert generate_ok.exit_code == 0

    generate_bad = runner.invoke(yaral, ["generate", bad_path])
    assert generate_bad.exit_code != 0

    compare_sem = runner.invoke(yaral, ["compare", good, good, "--semantic"])
    assert compare_sem.exit_code == 0

    compare_struct = runner.invoke(yaral, ["compare", good, other])
    assert compare_struct.exit_code == 0

    compare_bad = runner.invoke(yaral, ["compare", good, bad_path])
    assert compare_bad.exit_code != 0

    info = runner.invoke(yaral, ["info", "--examples", "--fields", "--functions"])
    assert info.exit_code == 0


def test_analyze_real_paths(tmp_path: Path) -> None:
    runner = CliRunner()

    good = _write(tmp_path, "ok.yar", _sample_yara())
    bad = _write(tmp_path, "bad.yar", _invalid_rule())

    full_json = runner.invoke(analyze, ["full", good, "--format", "json"])
    assert full_json.exit_code == 0
    assert '"best_practices"' in full_json.output
    assert '"optimization"' in full_json.output

    full_text = runner.invoke(analyze, ["full", good])
    assert full_text.exit_code == 0

    out_json = tmp_path / "analysis.json"
    full_json_out = runner.invoke(analyze, ["full", good, "--format", "json", "-o", str(out_json)])
    assert full_json_out.exit_code == 0
    assert out_json.exists()

    full_bad = runner.invoke(analyze, ["full", bad])
    assert full_bad.exit_code != 0

    bp_ok = runner.invoke(analyze, ["best-practices", good, "--category", "all"])
    assert bp_ok.exit_code == 0

    bp_bad = runner.invoke(analyze, ["best-practices", bad])
    assert bp_bad.exit_code != 0

    opt_text = runner.invoke(analyze, ["optimize", good])
    assert opt_text.exit_code == 0

    opt_json = runner.invoke(analyze, ["optimize", good, "--format", "json"])
    assert opt_json.exit_code == 0
    assert '"statistics"' in opt_json.output
    assert '"suggestions"' in opt_json.output

    opt_out = tmp_path / "opt-analysis.json"
    opt_json_out = runner.invoke(
        analyze, ["optimize", good, "--format", "json", "-o", str(opt_out)]
    )
    assert opt_json_out.exit_code == 0
    assert opt_out.exists()

    opt_bad = runner.invoke(analyze, ["optimize", bad])
    assert opt_bad.exit_code != 0



"""More tests for YARA-L optimizer (no mocks)."""

from __future__ import annotations

from yaraast.ast.expressions import BooleanLiteral
from yaraast.yaral.ast_nodes import (
    BinaryCondition,
    ConditionSection,
    EventExistsCondition,
    MatchSection,
    MatchVariable,
    TimeWindow,
    YaraLFile,
    YaraLRule,
)
from yaraast.yaral.optimizer import YaraLOptimizer


def test_optimizer_simplifies_boolean_and_time_window() -> None:
    condition = ConditionSection(
        expression=BinaryCondition(
            operator="and",
            left=EventExistsCondition(event="e"),
            right=BooleanLiteral(value=True),
        ),
    )
    match = MatchSection(variables=[MatchVariable(variable="e", time_window=TimeWindow(3600, "s"))])
    rule = YaraLRule(name="opt", events=None, condition=condition, match=match)
    ast = YaraLFile(rules=[rule])

    optimized, stats = YaraLOptimizer().optimize(ast)
    assert stats.conditions_simplified >= 1
    assert stats.time_windows_optimized >= 1

    optimized_match = optimized.rules[0].match
    assert optimized_match.variables[0].time_window.unit == "h"



"""Coverage wave for CLI service/reporting helpers (no mocks)."""

from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace

from rich.console import Console

from yaraast.cli import analyze_services as an
from yaraast.cli import bench_services as bs
from yaraast.cli import performance_check_reporting as pcr
from yaraast.cli import workspace_services as ws
from yaraast.performance.string_analyzer import StringPerformanceIssue


class _Suggestion:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def format(self) -> str:
        return f"{self.rule_name}:{self.message if hasattr(self, 'message') else self.description}"


def test_analyze_services_formatting_helpers() -> None:
    bp = SimpleNamespace(
        statistics={"rules": 2},
        suggestions=[
            _Suggestion(
                rule_name="r1",
                category="security",
                severity="error",
                message="bad",
                location={"line": 1},
            ),
            _Suggestion(
                rule_name="r2",
                category="style",
                severity="warning",
                message="meh",
                location={"line": 2},
            ),
        ],
        get_by_severity=lambda sev: [s for s in bp.suggestions if s.severity == sev],
    )
    opt = SimpleNamespace(
        statistics={"suggestions": 1},
        suggestions=[
            _Suggestion(
                rule_name="r1",
                optimization_type="dedup",
                impact="high",
                description="improve",
                code_before="a",
                code_after="b",
            ),
        ],
    )

    errors, warnings, info = an._get_severity_counts(bp)
    assert len(errors) == 1 and len(warnings) == 1 and len(info) == 0

    assert an._filter_suggestions(bp.suggestions, "all") == bp.suggestions
    assert len(an._filter_suggestions(bp.suggestions, "security")) == 1

    data = an._generate_json_report("sample.yar", bp, opt)
    assert data["file"] == "sample.yar"
    assert data["best_practices"]["statistics"] == {"rules": 2}
    assert data["optimization"]["suggestions"][0]["type"] == "dedup"

    text = an._generate_text_report("sample.yar", bp, opt)
    assert "BEST PRACTICES" in text
    assert "OPTIMIZATIONS" in text
    assert "SUMMARY" in text
    assert "Total issues:" in text

    assert an._get_level_style("high") == "red"
    assert an._get_level_style("medium") == "yellow"
    assert an._get_level_style("low") == "blue"
    assert an._get_level_style("other") == "white"


def test_bench_services_operations_and_summary() -> None:
    class _Result:
        def __init__(self, success: bool):
            self.success = success

    class _Bench:
        def benchmark_parsing(self, *_args):
            return _Result(True)

        def benchmark_codegen(self, *_args):
            return _Result(False)

        def benchmark_roundtrip(self, *_args):
            return [_Result(True)]

        def get_benchmark_summary(self):
            return {"total": 3}

    bench = _Bench()
    path = Path("/tmp/rule.yar")

    assert bs._determine_operations_to_run("all") == ["parse", "codegen", "roundtrip"]
    assert bs._determine_operations_to_run("roundtrip") == ["roundtrip"]
    assert bs._determine_operations_to_run("parse") == ["parse"]

    assert bs._run_single_operation(bench, path, "parse", 2).success is True
    assert bs._run_single_operation(bench, path, "codegen", 2).success is False
    assert bs._run_single_operation(bench, path, "roundtrip", 2).success is True
    assert bs._run_single_operation(bench, path, "unknown", 2) is None

    file_results = bs._run_benchmarks_for_single_file(bench, path, "all", 3)
    assert "parse" in file_results
    assert "roundtrip" in file_results
    assert "codegen" not in file_results

    all_results = bs._run_benchmarks_for_all_files(bench, [path], "all", 1)
    assert all_results[0]["file_name"] == "rule.yar"
    assert bs._get_benchmark_summary(bench)["total"] == 3


def test_workspace_services_formatters() -> None:
    dep = SimpleNamespace(
        export_dot=lambda: "digraph G {}",
        nodes={
            "a": SimpleNamespace(
                type="rule",
                dependencies={"b"},
                dependents=set(),
                metadata={"x": 1},
            ),
        },
    )
    file_result = SimpleNamespace(
        errors=["e1"],
        warnings=["w1"],
        type_errors=["t1"],
        analysis_results={"k": "v"},
    )
    report = SimpleNamespace(
        files_analyzed=1,
        total_rules=2,
        total_includes=0,
        total_imports=1,
        global_errors=["g1"],
        file_results={"/tmp/a.yar": file_result},
        statistics={
            "total_errors": 1,
            "total_warnings": 1,
            "total_type_errors": 1,
            "cycles": 0,
            "rule_name_conflicts": 0,
        },
        dependency_graph=dep,
    )

    txt = ws.format_workspace_report_text(report)
    assert "Workspace Analysis Report" in txt
    assert "Global Errors:" in txt
    assert "Type Errors:" in txt

    js = ws.format_workspace_report_json(report)
    assert '"global_errors"' in js

    dot = ws.format_workspace_graph(report, "dot")
    assert "digraph" in dot

    graph_json = ws.format_workspace_graph(report, "json")
    assert '"nodes"' in graph_json

    assert ws.format_workspace_output(report, "json") == js
    assert ws.format_workspace_output(report, "dot") == dot
    assert "Workspace Analysis Report" in ws.format_workspace_output(report, "text")


def test_performance_check_reporting_render() -> None:
    console = Console(record=True, width=120)
    issues = [
        StringPerformanceIssue(
            rule_name="r1",
            string_id="$a",
            issue_type="dup",
            severity="critical",
            description="duplicate",
            suggestion="merge",
        ),
        StringPerformanceIssue(
            rule_name="r2",
            string_id="$b",
            issue_type="slow_regex",
            severity="medium",
            description="slow",
            suggestion="anchor",
        ),
    ]

    pcr.display_parse_failure(console)
    pcr.display_no_issues(console)
    pcr.display_issues(console, issues)
    pcr.display_summary(
        console,
        issue_types={
            "dup": {"count": 1, "critical": 1, "rules": {"r1"}},
            "slow_regex": {"count": 1, "critical": 0, "rules": {"r2"}},
        },
        total_rules=4,
        issues=issues,
    )
    pcr.display_issue_totals(console, issues)

    out = console.export_text()
    assert "Failed to parse" in out
    assert "No performance issues found" in out
    assert "Performance Issues" in out
    assert "Suggestions:" in out
    assert "Performance Issue Summary" in out
    assert "Rules with issues" in out
    assert "critical issues" in out



"""
Basic smoke tests for YARA-L auxiliary modules.

Copyright (c) 2026 Marc Rivero López
Licensed under GPLv3. See LICENSE file for details.
This test suite validates real code behavior without mocks or stubs.
"""

from __future__ import annotations


# Test that modules can be imported
def test_import_yaral_modules() -> None:
    """Test that all YARA-L modules can be imported."""
    from yaraast.yaral import ast_nodes, lexer, parser, tokens

    assert ast_nodes is not None
    assert lexer is not None
    assert parser is not None
    assert tokens is not None


def test_import_yaral_enhanced_parser() -> None:
    """Test importing enhanced parser module."""
    from yaraast.yaral import enhanced_parser

    assert enhanced_parser is not None


def test_import_yaral_generator() -> None:
    """Test importing generator module."""
    from yaraast.yaral import generator

    assert generator is not None


def test_import_yaral_optimizer() -> None:
    """Test importing optimizer module."""
    from yaraast.yaral import optimizer

    assert optimizer is not None


def test_import_yaral_validator() -> None:
    """Test importing validator module."""
    from yaraast.yaral import validator

    assert validator is not None


# Test AST node creation
def test_create_ast_nodes() -> None:
    """Test creating YARA-L AST nodes."""
    from yaraast.yaral.ast_nodes import (
        EventVariable,
        MetaEntry,
        RegexPattern,
        TimeWindow,
        UDMFieldPath,
    )

    # Create instances
    event_var = EventVariable(name="$e")
    assert event_var.name == "$e"

    meta_entry = MetaEntry(key="author", value="Test")
    assert meta_entry.key == "author"
    assert meta_entry.value == "Test"

    time_window = TimeWindow(duration=5, unit="m")
    assert time_window.duration == 5
    assert time_window.unit == "m"
    assert time_window.as_string == "5m"

    field_path = UDMFieldPath(parts=["metadata", "event_type"])
    assert field_path.path == "metadata.event_type"

    regex = RegexPattern(pattern="test.*")
    assert regex.pattern == "test.*"


# Test token types
def test_yaral_token_types() -> None:
    """Test YARA-L token type enum."""
    from yaraast.yaral.tokens import YaraLTokenType

    # Verify key token types exist
    assert YaraLTokenType.EVENTS is not None
    assert YaraLTokenType.MATCH is not None
    assert YaraLTokenType.OUTCOME is not None
    assert YaraLTokenType.OVER is not None
    assert YaraLTokenType.EVENT_VAR is not None
    assert YaraLTokenType.REFERENCE_LIST is not None
    assert YaraLTokenType.TIME_LITERAL is not None


# Test lexer token creation
def test_yaral_token_creation() -> None:
    """Test creating YARA-L tokens."""
    from yaraast.lexer.tokens import TokenType as BaseTokenType
    from yaraast.yaral.lexer import YaraLToken
    from yaraast.yaral.tokens import YaraLTokenType

    token = YaraLToken(
        type=BaseTokenType.IDENTIFIER,
        value="test",
        line=1,
        column=1,
        length=4,
        yaral_type=YaraLTokenType.EVENTS,
    )

    assert token.value == "test"
    assert token.line == 1
    assert token.yaral_type == YaraLTokenType.EVENTS


# Test parser error class
def test_yaral_parser_error() -> None:
    """Test YARA-L parser error class."""
    from yaraast.lexer.tokens import TokenType as BaseTokenType
    from yaraast.yaral.lexer import YaraLToken
    from yaraast.yaral.parser import YaraLParserError

    # Create error without token
    error1 = YaraLParserError("Test error")
    assert "Test error" in str(error1)

    # Create error with token
    token = YaraLToken(type=BaseTokenType.IDENTIFIER, value="test", line=5, column=10, length=4)
    error2 = YaraLParserError("Test error", token)
    assert "Parser error at 5:10" in str(error2)
    assert error2.token == token


# Basic round-trip test
def test_parse_and_ast_properties() -> None:
    """Test parsing and accessing AST properties."""
    from yaraast.yaral.parser import YaraLParser

    yaral_code = """
    rule test_properties {
        meta:
            author = "Test"
            version = 1

        events:
            $e.metadata.event_type = "LOGIN"

        match:
            $hostname over 5m

        outcome:
            $count = count($e.metadata.id)

        condition:
            $count > 10

        options:
            enabled = true
    }
    """

    parser = YaraLParser(yaral_code)
    ast = parser.parse()

    # Verify all properties are accessible
    rule = ast.rules[0]
    assert rule.name == "test_properties"
    assert rule.meta is not None
    assert len(rule.meta.entries) == 2

    assert rule.events is not None
    assert len(rule.events.statements) > 0

    assert rule.match is not None
    assert len(rule.match.variables) > 0

    assert rule.outcome is not None
    assert len(rule.outcome.assignments) > 0

    assert rule.condition is not None
    assert rule.condition.expression is not None

    assert rule.options is not None
    assert len(rule.options.options) > 0


# Test file with multiple rules
def test_yaral_file_operations() -> None:
    """Test YaraLFile operations."""
    from yaraast.yaral.ast_nodes import EventsSection, YaraLFile, YaraLRule

    yaral_file = YaraLFile(rules=[])
    assert len(yaral_file.rules) == 0

    rule1 = YaraLRule(name="rule1", events=EventsSection())
    yaral_file.add_rule(rule1)
    assert len(yaral_file.rules) == 1

    rule2 = YaraLRule(name="rule2", events=EventsSection())
    yaral_file.add_rule(rule2)
    assert len(yaral_file.rules) == 2

    assert yaral_file.rules[0].name == "rule1"
    assert yaral_file.rules[1].name == "rule2"



"""Leaf quick-fix builders for semantic code actions."""

from __future__ import annotations

import re

from lsprotocol.types import (
    CodeAction,
    CodeActionKind,
    Diagnostic,
    Position,
    Range,
    TextEdit,
    WorkspaceEdit,
)


def create_replace_module_function_actions(
    text: str,
    diagnostic: Diagnostic,
    uri: str,
    module_name: str,
    function_name: str,
    available_functions: list[str],
) -> list[CodeAction]:
    actions: list[CodeAction] = []
    if not available_functions:
        return actions

    lines = text.split("\n")
    line_num = diagnostic.range.start.line
    if line_num >= len(lines):
        return actions
    line = lines[line_num]
    needle = f"{module_name}.{function_name}"
    start_col = line.find(needle)
    if start_col < 0:
        return actions

    for replacement_name in available_functions[:3]:
        replacement = f"{module_name}.{replacement_name}"
        actions.append(
            CodeAction(
                title=f"Replace with {replacement}",
                kind=CodeActionKind.QuickFix,
                edit=WorkspaceEdit(
                    changes={
                        uri: [
                            TextEdit(
                                range=Range(
                                    start=Position(line=line_num, character=start_col),
                                    end=Position(line=line_num, character=start_col + len(needle)),
                                ),
                                new_text=replacement,
                            )
                        ]
                    }
                ),
                diagnostics=[diagnostic],
            )
        )
    return actions


def create_replace_builtin_function_actions(
    text: str,
    diagnostic: Diagnostic,
    uri: str,
    function_name: str,
    suggested_functions: list[str],
) -> list[CodeAction]:
    actions: list[CodeAction] = []
    if not suggested_functions:
        return actions

    lines = text.split("\n")
    line_num = diagnostic.range.start.line
    if line_num >= len(lines):
        return actions

    line = lines[line_num]
    start_col = line.find(function_name)
    if start_col < 0:
        return actions

    for replacement_name in suggested_functions[:3]:
        actions.append(
            CodeAction(
                title=f"Replace with {replacement_name}()",
                kind=CodeActionKind.QuickFix,
                edit=WorkspaceEdit(
                    changes={
                        uri: [
                            TextEdit(
                                range=Range(
                                    start=Position(line=line_num, character=start_col),
                                    end=Position(
                                        line=line_num, character=start_col + len(function_name)
                                    ),
                                ),
                                new_text=replacement_name,
                            )
                        ]
                    }
                ),
                diagnostics=[diagnostic],
            )
        )

    return actions


def create_add_placeholder_argument_action(
    text: str,
    diagnostic: Diagnostic,
    uri: str,
    function_name: str,
) -> list[CodeAction]:
    lines = text.split("\n")
    line_num = diagnostic.range.start.line
    if line_num >= len(lines):
        return []
    line = lines[line_num]
    needle = f"{function_name}()"
    start_col = line.find(needle)
    if start_col < 0:
        return []

    edit = TextEdit(
        range=Range(
            start=Position(line=line_num, character=start_col + len(function_name) + 1),
            end=Position(line=line_num, character=start_col + len(function_name) + 1),
        ),
        new_text="0",
    )
    return [
        CodeAction(
            title=f"Add placeholder argument to {function_name}()",
            kind=CodeActionKind.QuickFix,
            edit=WorkspaceEdit(changes={uri: [edit]}),
            diagnostics=[diagnostic],
        )
    ]


def create_add_missing_arguments_action(
    text: str,
    diagnostic: Diagnostic,
    uri: str,
    function_name: str,
    missing_count: int,
) -> list[CodeAction]:
    if missing_count <= 0:
        return []
    lines = text.split("\n")
    line_num = diagnostic.range.start.line
    if line_num >= len(lines):
        return []
    line = lines[line_num]
    start_col = line.find(f"{function_name}(")
    if start_col < 0:
        return []
    open_paren = start_col + len(function_name)
    close_paren = line.find(")", open_paren)
    if close_paren < 0:
        return []
    insertion = ("0, " * missing_count).rstrip(", ")
    if close_paren > open_paren + 1:
        insertion = ", " + insertion
    return [
        CodeAction(
            title=f"Add {missing_count} missing argument(s) to {function_name}()",
            kind=CodeActionKind.QuickFix,
            edit=WorkspaceEdit(
                changes={
                    uri: [
                        TextEdit(
                            range=Range(
                                start=Position(line=line_num, character=close_paren),
                                end=Position(line=line_num, character=close_paren),
                            ),
                            new_text=insertion,
                        )
                    ]
                }
            ),
            diagnostics=[diagnostic],
        )
    ]


def create_trim_arguments_action(
    text: str,
    diagnostic: Diagnostic,
    uri: str,
    function_name: str,
    keep_args: int,
) -> list[CodeAction]:
    lines = text.split("\n")
    line_num = diagnostic.range.start.line
    if line_num >= len(lines) or keep_args < 0:
        return []
    line = lines[line_num]
    start_col = line.find(f"{function_name}(")
    if start_col < 0:
        return []
    open_paren = start_col + len(function_name)
    close_paren = line.find(")", open_paren)
    if close_paren < 0:
        return []

    args_text = line[open_paren + 1 : close_paren]
    parts = [part.strip() for part in args_text.split(",")]
    if len(parts) <= keep_args:
        return []
    replacement = ", ".join(parts[:keep_args])

    return [
        CodeAction(
            title=f"Remove extra argument(s) from {function_name}()",
            kind=CodeActionKind.QuickFix,
            edit=WorkspaceEdit(
                changes={
                    uri: [
                        TextEdit(
                            range=Range(
                                start=Position(line=line_num, character=open_paren + 1),
                                end=Position(line=line_num, character=close_paren),
                            ),
                            new_text=replacement,
                        )
                    ]
                }
            ),
            diagnostics=[diagnostic],
        )
    ]


def create_import_module_action(
    module_name: str, diagnostic: Diagnostic, uri: str
) -> list[CodeAction]:
    import_line = f'import "{module_name}"\n'
    insert_position = Position(line=0, character=0)
    edit = TextEdit(range=Range(start=insert_position, end=insert_position), new_text=import_line)
    return [
        CodeAction(
            title=f'Add import "{module_name}"',
            kind=CodeActionKind.QuickFix,
            edit=WorkspaceEdit(changes={uri: [edit]}),
            diagnostics=[diagnostic],
        )
    ]


def create_rename_duplicate_action(
    text: str,
    diagnostic: Diagnostic,
    uri: str,
    identifier: str,
) -> list[CodeAction]:
    base_name = identifier.removeprefix("$")
    if not base_name:
        return []

    lines = text.split("\n")
    existing_ids = set()
    for line in lines:
        id_match = re.search(r"\$(\w+)\s*=", line)
        if id_match:
            existing_ids.add(id_match.group(1))

    counter = 2
    while f"{base_name}_{counter}" in existing_ids:
        counter += 1
    new_name = f"${base_name}_{counter}"

    line_num = diagnostic.range.start.line
    if line_num < len(lines):
        line = lines[line_num]
        col = line.find(f"${base_name}")
        if col >= 0:
            edit = TextEdit(
                range=Range(
                    start=Position(line=line_num, character=col),
                    end=Position(line=line_num, character=col + len(f"${base_name}")),
                ),
                new_text=new_name,
            )
            return [
                CodeAction(
                    title=f"Rename to {new_name}",
                    kind=CodeActionKind.QuickFix,
                    edit=WorkspaceEdit(changes={uri: [edit]}),
                    diagnostics=[diagnostic],
                )
            ]
    return []



"""Extra tests for Protobuf serializer (no mocks)."""

from __future__ import annotations

from textwrap import dedent

from yaraast.parser import Parser
from yaraast.serialization.protobuf_serializer import ProtobufSerializer


def test_protobuf_hex_regex_and_deserialize_file(tmp_path) -> None:
    code = dedent(
        """
        rule proto2 {
            strings:
                $a = { 4D 5A ?? [2-3] 00 }
                $b = /test.*/i
            condition:
                $a and $b
        }
        """,
    )
    ast = Parser().parse(code)
    serializer = ProtobufSerializer(include_metadata=True)

    out_path = tmp_path / "ast2.pb"
    serializer.serialize(ast, output_path=out_path)

    restored = serializer.deserialize(input_path=out_path)
    assert len(restored.rules) == 1
    assert restored.rules[0].name == "proto2"

    stats = serializer.get_serialization_stats(ast)
    assert stats["binary_size_bytes"] > 0



"""Operator-specific AST nodes."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from yaraast.ast.base import _VisitorType
from yaraast.ast.expressions import Expression


@dataclass
class DefinedExpression(Expression):
    """Defined operator expression."""

    expression: Expression

    def accept(self, visitor: _VisitorType) -> Any:
        return visitor.visit_defined_expression(self)


@dataclass
class StringOperatorExpression(Expression):
    """String comparison operator expression (iequals, icontains, etc.)."""

    left: Expression
    operator: str  # "iequals", "icontains", "istartswith", "iendswith"
    right: Expression

    def accept(self, visitor: _VisitorType) -> Any:
        return visitor.visit_string_operator_expression(self)



"""Additional tests for performance-check service helpers."""

from __future__ import annotations

from yaraast.cli.performance_check_services import filter_issues, summarize_issues
from yaraast.performance.string_analyzer import StringPerformanceIssue


def _issue(rule_name: str, issue_type: str, severity: str) -> StringPerformanceIssue:
    return StringPerformanceIssue(
        rule_name=rule_name,
        string_id="$a",
        issue_type=issue_type,
        severity=severity,
        description="desc",
        suggestion="fix",
    )


def test_filter_issues_applies_warning_filter_and_limit() -> None:
    issues = [
        _issue("r1", "short_string", "warning"),
        _issue("r2", "short_string", "critical"),
        _issue("r3", "slow_regex", "warning"),
    ]

    filtered = filter_issues(issues, "warning", 1)

    assert len(filtered) == 1
    assert filtered[0].severity == "warning"
    assert filtered[0].rule_name == "r1"


def test_summarize_issues_counts_criticals_and_rules() -> None:
    issues = [
        _issue("r1", "short_string", "critical"),
        _issue("r2", "short_string", "warning"),
        _issue("r1", "short_string", "critical"),
        _issue("r3", "slow_regex", "warning"),
    ]

    summary = summarize_issues(issues)

    assert summary["short_string"]["count"] == 3
    assert summary["short_string"]["critical"] == 2
    assert summary["short_string"]["rules"] == {"r1", "r2"}
    assert summary["slow_regex"]["count"] == 1
    assert summary["slow_regex"]["critical"] == 0



"""Reporting helpers for tag diffs."""

from __future__ import annotations


def emit_tags_diff(
    base_path: str,
    result,
    diff_node,
    diff_type,
    old_tags: set,
    new_tags: set,
) -> None:
    """Record tags diff."""
    result.differences.append(
        diff_node(
            path=f"{base_path}/tags",
            diff_type=diff_type.MODIFIED,
            old_value=list(old_tags),
            new_value=list(new_tags),
            node_type="RuleTags",
        ),
    )



"""Shared LSP types/constants with robust import handling."""

from __future__ import annotations

try:
    from lsprotocol.types import (  # type: ignore
        TEXT_DOCUMENT_CODE_ACTION,
        TEXT_DOCUMENT_COMPLETION,
        TEXT_DOCUMENT_DEFINITION,
        TEXT_DOCUMENT_DIAGNOSTIC,
        TEXT_DOCUMENT_DID_CHANGE,
        TEXT_DOCUMENT_DID_CLOSE,
        TEXT_DOCUMENT_DID_OPEN,
        TEXT_DOCUMENT_DID_SAVE,
        TEXT_DOCUMENT_DOCUMENT_HIGHLIGHT,
        TEXT_DOCUMENT_DOCUMENT_LINK,
        TEXT_DOCUMENT_DOCUMENT_SYMBOL,
        TEXT_DOCUMENT_FOLDING_RANGE,
        TEXT_DOCUMENT_FORMATTING,
        TEXT_DOCUMENT_HOVER,
        TEXT_DOCUMENT_PREPARE_RENAME,
        TEXT_DOCUMENT_RANGE_FORMATTING,
        TEXT_DOCUMENT_REFERENCES,
        TEXT_DOCUMENT_RENAME,
        TEXT_DOCUMENT_SELECTION_RANGE,
        TEXT_DOCUMENT_SEMANTIC_TOKENS_FULL,
        TEXT_DOCUMENT_SEMANTIC_TOKENS_RANGE,
        TEXT_DOCUMENT_SIGNATURE_HELP,
        WORKSPACE_DID_CHANGE_CONFIGURATION,
        WORKSPACE_DID_CHANGE_WATCHED_FILES,
        WORKSPACE_SYMBOL,
        CodeAction,
        CodeActionOptions,
        CodeActionParams,
        CompletionList,
        CompletionOptions,
        CompletionParams,
        DefinitionParams,
        DiagnosticOptions,
        DidChangeConfigurationParams,
        DidChangeTextDocumentParams,
        DidChangeWatchedFilesParams,
        DidCloseTextDocumentParams,
        DidOpenTextDocumentParams,
        DidSaveTextDocumentParams,
        DocumentDiagnosticParams,
        DocumentDiagnosticReportKind,
        DocumentFormattingParams,
        DocumentHighlight,
        DocumentHighlightParams,
        DocumentLink,
        DocumentLinkParams,
        DocumentRangeFormattingParams,
        DocumentSymbol,
        DocumentSymbolParams,
        FoldingRange,
        FoldingRangeParams,
        FullDocumentDiagnosticReport,
        Hover,
        HoverParams,
        InitializeParams,
        Location,
        PrepareRenameParams,
        Range,
        ReferenceParams,
        RelatedFullDocumentDiagnosticReport,
        RenameParams,
        SelectionRange,
        SelectionRangeParams,
        SemanticTokens,
        SemanticTokensParams,
        SemanticTokensRangeParams,
        SemanticTokensRegistrationOptions,
        SignatureHelp,
        SignatureHelpOptions,
        SignatureHelpParams,
        SymbolInformation,
        TextEdit,
        WorkspaceEdit,
        WorkspaceFoldersServerCapabilities,
        WorkspaceSymbolParams,
    )
except ImportError:  # pragma: no cover - exercised in environments without lsprotocol
    import importlib
    import site
    import sys

    for path in reversed(site.getsitepackages()):
        if path in sys.path:
            sys.path.remove(path)
        sys.path.insert(0, path)
    sys.modules.pop("lsprotocol", None)
    sys.modules.pop("lsprotocol.types", None)
    importlib.invalidate_caches()
    from lsprotocol.types import (  # type: ignore
        TEXT_DOCUMENT_CODE_ACTION,
        TEXT_DOCUMENT_COMPLETION,
        TEXT_DOCUMENT_DEFINITION,
        TEXT_DOCUMENT_DIAGNOSTIC,
        TEXT_DOCUMENT_DID_CHANGE,
        TEXT_DOCUMENT_DID_CLOSE,
        TEXT_DOCUMENT_DID_OPEN,
        TEXT_DOCUMENT_DID_SAVE,
        TEXT_DOCUMENT_DOCUMENT_HIGHLIGHT,
        TEXT_DOCUMENT_DOCUMENT_LINK,
        TEXT_DOCUMENT_DOCUMENT_SYMBOL,
        TEXT_DOCUMENT_FOLDING_RANGE,
        TEXT_DOCUMENT_FORMATTING,
        TEXT_DOCUMENT_HOVER,
        TEXT_DOCUMENT_PREPARE_RENAME,
        TEXT_DOCUMENT_RANGE_FORMATTING,
        TEXT_DOCUMENT_REFERENCES,
        TEXT_DOCUMENT_RENAME,
        TEXT_DOCUMENT_SELECTION_RANGE,
        TEXT_DOCUMENT_SEMANTIC_TOKENS_FULL,
        TEXT_DOCUMENT_SEMANTIC_TOKENS_RANGE,
        TEXT_DOCUMENT_SIGNATURE_HELP,
        WORKSPACE_DID_CHANGE_CONFIGURATION,
        WORKSPACE_DID_CHANGE_WATCHED_FILES,
        WORKSPACE_SYMBOL,
        CodeAction,
        CodeActionOptions,
        CodeActionParams,
        CompletionList,
        CompletionOptions,
        CompletionParams,
        DefinitionParams,
        DiagnosticOptions,
        DidChangeConfigurationParams,
        DidChangeTextDocumentParams,
        DidChangeWatchedFilesParams,
        DidCloseTextDocumentParams,
        DidOpenTextDocumentParams,
        DidSaveTextDocumentParams,
        DocumentDiagnosticParams,
        DocumentDiagnosticReportKind,
        DocumentFormattingParams,
        DocumentHighlight,
        DocumentHighlightParams,
        DocumentLink,
        DocumentLinkParams,
        DocumentRangeFormattingParams,
        DocumentSymbol,
        DocumentSymbolParams,
        FoldingRange,
        FoldingRangeParams,
        FullDocumentDiagnosticReport,
        Hover,
        HoverParams,
        InitializeParams,
        Location,
        PrepareRenameParams,
        Range,
        ReferenceParams,
        RelatedFullDocumentDiagnosticReport,
        RenameParams,
        SelectionRange,
        SelectionRangeParams,
        SemanticTokens,
        SemanticTokensParams,
        SemanticTokensRangeParams,
        SemanticTokensRegistrationOptions,
        SignatureHelp,
        SignatureHelpOptions,
        SignatureHelpParams,
        SymbolInformation,
        TextEdit,
        WorkspaceEdit,
        WorkspaceFoldersServerCapabilities,
        WorkspaceSymbolParams,
    )

YARAAST_RUNTIME_STATUS = "yaraast/status"

__all__ = [
    "TEXT_DOCUMENT_CODE_ACTION",
    "TEXT_DOCUMENT_COMPLETION",
    "TEXT_DOCUMENT_DEFINITION",
    "TEXT_DOCUMENT_DIAGNOSTIC",
    "TEXT_DOCUMENT_DID_CHANGE",
    "TEXT_DOCUMENT_DID_CLOSE",
    "TEXT_DOCUMENT_DID_OPEN",
    "TEXT_DOCUMENT_DID_SAVE",
    "TEXT_DOCUMENT_DOCUMENT_HIGHLIGHT",
    "TEXT_DOCUMENT_DOCUMENT_LINK",
    "TEXT_DOCUMENT_DOCUMENT_SYMBOL",
    "TEXT_DOCUMENT_FOLDING_RANGE",
    "TEXT_DOCUMENT_FORMATTING",
    "TEXT_DOCUMENT_HOVER",
    "TEXT_DOCUMENT_PREPARE_RENAME",
    "TEXT_DOCUMENT_RANGE_FORMATTING",
    "TEXT_DOCUMENT_REFERENCES",
    "TEXT_DOCUMENT_RENAME",
    "TEXT_DOCUMENT_SELECTION_RANGE",
    "TEXT_DOCUMENT_SEMANTIC_TOKENS_FULL",
    "TEXT_DOCUMENT_SEMANTIC_TOKENS_RANGE",
    "TEXT_DOCUMENT_SIGNATURE_HELP",
    "WORKSPACE_DID_CHANGE_CONFIGURATION",
    "WORKSPACE_DID_CHANGE_WATCHED_FILES",
    "WORKSPACE_SYMBOL",
    "YARAAST_RUNTIME_STATUS",
    "CodeAction",
    "CodeActionOptions",
    "CodeActionParams",
    "CompletionList",
    "CompletionOptions",
    "CompletionParams",
    "DefinitionParams",
    "DiagnosticOptions",
    "DidChangeConfigurationParams",
    "DidChangeTextDocumentParams",
    "DidChangeWatchedFilesParams",
    "DidCloseTextDocumentParams",
    "DidOpenTextDocumentParams",
    "DidSaveTextDocumentParams",
    "DocumentDiagnosticParams",
    "DocumentDiagnosticReportKind",
    "DocumentFormattingParams",
    "DocumentHighlight",
    "DocumentHighlightParams",
    "DocumentLink",
    "DocumentLinkParams",
    "DocumentRangeFormattingParams",
    "DocumentSymbol",
    "DocumentSymbolParams",
    "FoldingRange",
    "FoldingRangeParams",
    "FullDocumentDiagnosticReport",
    "Hover",
    "HoverParams",
    "InitializeParams",
    "Location",
    "PrepareRenameParams",
    "Range",
    "ReferenceParams",
    "RelatedFullDocumentDiagnosticReport",
    "RenameParams",
    "SelectionRange",
    "SelectionRangeParams",
    "SemanticTokens",
    "SemanticTokensParams",
    "SemanticTokensRangeParams",
    "SemanticTokensRegistrationOptions",
    "SignatureHelp",
    "SignatureHelpOptions",
    "SignatureHelpParams",
    "SymbolInformation",
    "TextEdit",
    "WorkspaceEdit",
    "WorkspaceFoldersServerCapabilities",
    "WorkspaceSymbolParams",
]



"""Flow helpers for the error-tolerant parser."""

from __future__ import annotations

import re

from yaraast.ast.base import Location, YaraFile
from yaraast.ast.rules import Rule


def parse_with_recovery(parser) -> YaraFile:
    """Recover a YARA file by scanning line by line."""
    yara_file = YaraFile()
    yara_file.imports = []
    yara_file.includes = []
    yara_file.rules = []

    line_index = 0
    while line_index < len(parser.lines):
        line = parser.lines[line_index].strip()

        if not line or line.startswith(("//", "/*")):
            line_index += 1
            continue

        if line.startswith("import "):
            import_stmt = parser._parse_import_line(line, line_index)
            if import_stmt:
                yara_file.imports.append(import_stmt)
            line_index += 1
            continue

        if line.startswith("include "):
            include_stmt = parser._parse_include_line(line, line_index)
            if include_stmt:
                yara_file.includes.append(include_stmt)
            line_index += 1
            continue

        if line.startswith("rule ") or line.startswith("private ") or line.startswith("global "):
            rule, lines_consumed = parse_rule_with_recovery(parser, line_index)
            if rule:
                yara_file.rules.append(rule)
                parser.recovered_rules.append(rule)
            line_index += lines_consumed
            continue

        parser._add_error(f"Unexpected line: {line}", line_index, 0)
        line_index += 1

    _set_yara_file_location(yara_file)
    return yara_file


def parse_rule_with_recovery(parser, start_line: int) -> tuple[Rule | None, int]:
    """Recover a single rule block."""
    line = parser.lines[start_line].strip()
    rule_name, tags, modifiers = extract_rule_header(parser, line, start_line)
    if not rule_name:
        return None, 1

    rule_body_lines, current_line = collect_rule_body(parser, start_line, line)
    rule = parser._create_rule_from_body(
        rule_name, tags, body_lines=rule_body_lines, start_line=start_line
    )
    if rule:
        rule.modifiers = modifiers
    return rule, current_line - start_line + 1


def extract_rule_header(parser, line: str, line_num: int) -> tuple[str | None, list, list[str]]:
    """Extract rule name and tags from a rule declaration."""
    match = re.match(r"(?:(?:private|global)\s+)*rule\s+(\w+)\s*(?:\:\s*([^{]+))?\s*\{?", line)
    if not match:
        parser._add_error(f"Invalid rule declaration: {line}", line_num, 0)
        return None, [], []

    rule_name = match.group(1)
    tags_str = match.group(2)
    tags = [tag.strip() for tag in tags_str.split()] if tags_str else []
    return rule_name, tags, []


def _count_braces_outside_literals(line: str) -> int:
    """Count net brace balance ignoring braces inside strings, line comments, and block comments."""
    count = 0
    in_string = False
    in_block_comment = False
    i = 0
    while i < len(line):
        ch = line[i]
        if in_block_comment:
            if ch == "*" and i + 1 < len(line) and line[i + 1] == "/":
                in_block_comment = False
                i += 2
                continue
        elif in_string:
            if ch == "\\" and i + 1 < len(line):
                i += 2
                continue
            if ch == '"':
                in_string = False
        elif ch == '"':
            in_string = True
        elif ch == "/" and i + 1 < len(line) and line[i + 1] == "/":
            break  # Rest of line is a line comment
        elif ch == "/" and i + 1 < len(line) and line[i + 1] == "*":
            in_block_comment = True
            i += 2
            continue
        elif ch == "{":
            count += 1
        elif ch == "}":
            count -= 1
        i += 1
    return count


def collect_rule_body(parser, start_line: int, header_line: str) -> tuple[list[str], int]:
    """Collect all lines belonging to a rule body."""
    rule_body_lines: list[str] = []
    brace_count = _count_braces_outside_literals(header_line)
    current_line = start_line + 1

    while current_line < len(parser.lines) and (
        brace_count > 0 or (brace_count == 0 and not rule_body_lines)
    ):
        body_line = parser.lines[current_line]
        rule_body_lines.append(body_line)

        brace_count += _count_braces_outside_literals(body_line)
        if brace_count == 0:
            break
        current_line += 1

    return rule_body_lines, current_line


def _set_yara_file_location(yara_file: YaraFile) -> None:
    if not (yara_file.imports or yara_file.includes or yara_file.rules):
        return
    start_node = (
        yara_file.imports[0]
        if yara_file.imports
        else yara_file.includes[0]
        if yara_file.includes
        else yara_file.rules[0]
    )
    end_node = (
        yara_file.rules[-1]
        if yara_file.rules
        else yara_file.includes[-1]
        if yara_file.includes
        else yara_file.imports[-1]
    )
    start = start_node.location
    end = end_node.location
    if start is None or end is None:
        return
    yara_file.location = Location(
        line=start.line,
        column=start.column,
        end_line=end.end_line or end.line,
        end_column=end.end_column or (end.column + 1),
    )



"""Tests for small codegen formatting helper modules."""

from __future__ import annotations

from types import SimpleNamespace

from yaraast.ast.base import YaraFile
from yaraast.ast.expressions import BooleanLiteral, Identifier
from yaraast.ast.modifiers import StringModifier
from yaraast.ast.rules import Rule, Tag
from yaraast.ast.strings import HexByte, HexJump, HexString, PlainString, RegexString
from yaraast.codegen.generator_formatting import (
    escape_string_literal,
    format_boolean_literal,
    format_hex_jump,
    format_meta_value,
    format_regex_literal,
    format_rule_modifiers,
    format_rule_tags,
)
from yaraast.codegen.pretty_printer_helpers import (
    build_hex_pattern,
    calculate_meta_alignment_column,
    calculate_string_alignment_column,
    expression_to_string,
    format_plain_string,
    format_regex_string,
    modifiers_to_string,
)


def test_generator_formatting_helpers_cover_all_branches() -> None:
    assert format_rule_modifiers([]) == ""
    assert format_rule_modifiers(["private", "global"]) == "private global"
    assert format_rule_modifiers("private") == ""  # bare strings no longer accepted
    assert format_rule_modifiers(123) == ""

    assert format_rule_tags([]) == ""
    assert format_rule_tags(["t1", Tag(name="t2")]) == "t1 t2"

    assert format_meta_value("s", "x") == 's = "x"'
    assert format_meta_value("b", True) == "b = true"
    assert format_meta_value("n", 7) == "n = 7"

    assert escape_string_literal('a\\"b') == 'a\\\\\\"b'
    assert format_regex_literal("ab+", "is") == "/ab+/is"
    assert format_boolean_literal(True) == "true"
    assert format_boolean_literal(False) == "false"

    assert format_hex_jump(None, None) == "[-]"
    assert format_hex_jump(3, 3) == "[3]"
    assert format_hex_jump(None, 5) == "[-5]"
    assert format_hex_jump(2, None) == "[2-]"
    assert format_hex_jump(2, 5) == "[2-5]"


def test_pretty_printer_helpers_cover_all_branches() -> None:
    hex_node = HexString(
        "$h",
        tokens=[HexByte(0x4D), HexByte("5a"), HexJump(2, 2), HexJump(1, 3), SimpleNamespace()],
    )
    assert build_hex_pattern(hex_node, hex_uppercase=True, hex_spacing=True) == "4D 5A [2] [1-3] ??"
    assert build_hex_pattern(hex_node, hex_uppercase=False, hex_spacing=False) == "4d5a[2][1-3]??"

    plain = PlainString("$a", value="hello")
    regex = RegexString("$r", regex="ab+")
    assert format_plain_string(plain, '"', 3) == '$a    = "hello"'
    assert format_plain_string(plain, '"', 0) == '$a = "hello"'
    assert format_regex_string(regex, 2) == "$r   = /ab+/"
    assert format_regex_string(regex, 0) == "$r = /ab+/"

    assert modifiers_to_string([]) == ""
    assert (
        modifiers_to_string(
            [StringModifier.from_name_value("ascii"), StringModifier.from_name_value("nocase")]
        )
        == " ascii nocase"
    )

    ast = YaraFile(
        rules=[
            Rule(
                name="r",
                meta={"author": "me", "description": "x"},
                strings=[plain, HexString("$long_identifier", tokens=[]), regex],
                condition=BooleanLiteral(True),
            ),
            Rule(name="r2", meta=[], strings=[], condition=BooleanLiteral(True)),
        ]
    )
    assert calculate_string_alignment_column(ast) == len("$long_identifier") + 1
    # max() ensures alignment is at least min_alignment_column OR wide enough for longest key
    assert calculate_meta_alignment_column(ast, min_alignment_column=50) == 50  # min_alignment wins
    assert (
        calculate_meta_alignment_column(ast, min_alignment_column=5) == len("description =") + 2
    )  # key length wins

    assert expression_to_string(BooleanLiteral(True)) == "true"
    assert expression_to_string(Identifier("abc")) == "abc"



"""Rule resolution and reference queries for the LSP runtime."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from lsprotocol.types import Location, Position, TextEdit

from yaraast.lsp.document_types import ReferenceRecord, ResolvedSymbol, RuleLinkRecord, uri_to_path

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from yaraast.lsp.runtime import LspRuntime


def resolve_symbol(
    runtime: LspRuntime, uri: str, text: str, position: Position
) -> ResolvedSymbol | None:
    resolved = runtime.ensure_document(uri, text).resolve_symbol(position)
    if (
        resolved is not None
        and resolved.kind == "identifier"
        and find_rule_definition(runtime, resolved.normalized_name, uri) is not None
    ):
        return ResolvedSymbol(
            uri=resolved.uri,
            name=resolved.name,
            normalized_name=resolved.normalized_name,
            kind="rule",
            range=resolved.range,
        )
    return resolved


def find_rule_definition(
    runtime: LspRuntime,
    rule_name: str,
    current_uri: str | None = None,
) -> Location | None:
    cache_key = (runtime.cache.generation, rule_name, current_uri)
    if cache_key in runtime.cache.rule_definition_cache:
        return runtime.cache.rule_definition_cache[cache_key]
    ordered_docs = runtime.iter_workspace_documents()
    if current_uri:
        ordered_docs.sort(key=lambda doc: doc.uri != current_uri)
    for doc in ordered_docs:
        location = doc.find_rule_definition(rule_name)
        if location is not None:
            runtime.cache.rule_definition_cache[cache_key] = location
            return location
    runtime.cache.rule_definition_cache[cache_key] = None
    return None


def find_rule_references(
    runtime: LspRuntime,
    rule_name: str,
    *,
    include_declaration: bool = True,
    current_uri: str | None = None,
) -> list[Location]:
    cache_key = (runtime.cache.generation, rule_name, include_declaration, current_uri)
    cached = runtime.cache.rule_references_cache.get(cache_key)
    if cached is not None:
        return cached
    refs: list[Location] = []
    definition = find_rule_definition(runtime, rule_name, current_uri)
    for doc in runtime.iter_workspace_documents():
        refs.extend(doc.rule_occurrences(rule_name))
    if not include_declaration and definition is not None:
        refs = [
            ref for ref in refs if not (ref.uri == definition.uri and ref.range == definition.range)
        ]
    runtime.cache.rule_references_cache[cache_key] = refs
    return refs


def find_rule_reference_records(
    runtime: LspRuntime,
    rule_name: str,
    *,
    include_declaration: bool = True,
    current_uri: str | None = None,
) -> list[ReferenceRecord]:
    cache_key = (runtime.cache.generation, rule_name, include_declaration, current_uri)
    cached = runtime.cache.rule_reference_records_cache.get(cache_key)
    if cached is not None:
        return list(cached)
    records: list[ReferenceRecord] = []
    definition = find_rule_definition(runtime, rule_name, current_uri)
    for doc in runtime.iter_workspace_documents():
        for record in doc.rule_reference_records(
            rule_name, include_declaration=include_declaration
        ):
            if (
                not include_declaration
                and definition
                and (
                    record.location.uri == definition.uri
                    and record.location.range == definition.range
                )
            ):
                continue
            if (
                definition
                and record.location.uri == definition.uri
                and record.location.range == definition.range
            ):
                record = ReferenceRecord(record.location, "declaration", "rule")
            records.append(record)
    runtime.cache.rule_reference_records_cache[cache_key] = records
    return records


def find_rule_reference_records_in_document(
    runtime: LspRuntime,
    rule_name: str,
    document_uri: str,
    *,
    include_declaration: bool = True,
    current_uri: str | None = None,
) -> list[ReferenceRecord]:
    doc = runtime.documents.get(document_uri)
    if doc is None:
        path = uri_to_path(document_uri)
        if path is None or not path.exists() or not path.is_file():
            return []
        try:
            doc = runtime.get_document(document_uri)
        except Exception:
            logger.debug("Operation failed in %s", __name__, exc_info=True)
            return []
    if doc is None:
        return []
    definition = find_rule_definition(runtime, rule_name, current_uri)
    records: list[ReferenceRecord] = []
    for record in doc.rule_reference_records(rule_name, include_declaration=True):
        if (
            not include_declaration
            and definition
            and (
                record.location.uri == definition.uri and record.location.range == definition.range
            )
        ):
            continue
        if (
            definition
            and record.location.uri == definition.uri
            and record.location.range == definition.range
        ):
            record = ReferenceRecord(record.location, "declaration", "rule")
        records.append(record)
    return records


def get_rule_link_records_for_document(
    runtime: LspRuntime, document_uri: str
) -> list[RuleLinkRecord]:
    doc = runtime.documents.get(document_uri)
    if doc is None:
        path = uri_to_path(document_uri)
        if path is None or not path.exists() or not path.is_file():
            return []
        try:
            doc = runtime.get_document(document_uri)
        except Exception:
            logger.debug("Operation failed in %s", __name__, exc_info=True)
            return []
    workspace_rule_names = {
        record.name for record in runtime.workspace_symbol_records() if record.kind == "rule"
    }
    cache_key = f"rule_link_records:{runtime.cache.generation}:{document_uri}"
    cached = doc.get_cached(cache_key)
    if cached is not None:
        return cached
    links: list[RuleLinkRecord] = []
    for rule_name in workspace_rule_names:
        if not rule_name:
            continue
        definition = find_rule_definition(runtime, rule_name, document_uri)
        if definition is None:
            continue
        for record in find_rule_reference_records_in_document(
            runtime,
            rule_name,
            document_uri,
            current_uri=document_uri,
        ):
            if record.role == "declaration" or record.location.uri != document_uri:
                continue
            links.append(
                RuleLinkRecord(
                    rule_name=rule_name,
                    location=record.location,
                    target_uri=definition.uri,
                )
            )
    doc.set_cached(cache_key, links)
    return links


def rename_rule(runtime: LspRuntime, rule_name: str, new_name: str) -> dict[str, list[TextEdit]]:
    changes: dict[str, list[TextEdit]] = {}
    for doc in runtime.iter_workspace_documents():
        edits = doc.rename_rule_edits(rule_name, new_name)
        if edits:
            changes[doc.uri] = edits
    return changes



"""Rule-oriented query helpers for LSP document contexts."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from yaraast.lsp.document_context import DocumentContext


def get_rule_info(ctx: DocumentContext, rule_name: str) -> dict[str, Any] | None:
    cache_key = f"rule_info:{rule_name}"
    cached = ctx.get_cached(cache_key)
    if cached is not None:
        return cached
    rule = ctx.get_rule(rule_name)
    if rule is None:
        return None
    modifiers = getattr(rule, "modifiers", None) or []
    tags = getattr(rule, "tags", None) or []
    meta = getattr(rule, "meta", None)
    if isinstance(meta, list):
        meta_items = [(getattr(m, "key", ""), getattr(m, "value", "")) for m in meta]
    elif hasattr(meta, "entries"):
        meta_items = [(entry.key, entry.value) for entry in getattr(meta, "entries", [])]
    else:
        meta_items = []
    result = {
        "name": rule_name,
        "modifiers": [str(m) for m in modifiers],
        "tags": [tag.name if hasattr(tag, "name") else str(tag) for tag in tags],
        "meta": meta_items,
        "strings_count": len(getattr(rule, "strings", None) or []),
        "has_events": getattr(rule, "events", None) is not None,
        "has_match": getattr(rule, "match", None) is not None,
        "has_outcome": getattr(rule, "outcome", None) is not None,
        "has_options": getattr(rule, "options", None) is not None,
    }
    ctx.set_cached(cache_key, result)
    return result


def get_rule_meta_items(ctx: DocumentContext, rule_name: str) -> list[tuple[str, Any]]:
    cache_key = f"rule_meta_items:{rule_name}"
    cached = ctx.get_cached(cache_key)
    if cached is not None:
        return cached
    rule = ctx.get_rule(rule_name)
    if rule is None:
        return []
    meta = getattr(rule, "meta", None)
    if isinstance(meta, list):
        result = [(getattr(m, "key", ""), getattr(m, "value", "")) for m in meta]
    elif hasattr(meta, "entries"):
        result = [(entry.key, entry.value) for entry in getattr(meta, "entries", [])]
    else:
        result = []
    ctx.set_cached(cache_key, result)
    return result


def get_rule_string_identifiers(ctx: DocumentContext, rule_name: str) -> list[str]:
    cache_key = f"rule_string_identifiers:{rule_name}"
    cached = ctx.get_cached(cache_key)
    if cached is not None:
        return cached
    rule = ctx.get_rule(rule_name)
    if rule is None:
        return []
    result = [
        string_def.identifier
        for string_def in getattr(rule, "strings", []) or []
        if getattr(string_def, "identifier", None)
    ]
    ctx.set_cached(cache_key, result)
    return result


def get_rule_sections(ctx: DocumentContext, rule_name: str) -> list[str]:
    cache_key = f"rule_sections:{rule_name}"
    cached = ctx.get_cached(cache_key)
    if cached is not None:
        return cached
    sections: list[str] = []
    seen: set[str] = set()
    for symbol in ctx._symbols_of_kind("section"):
        if symbol.container_name == rule_name and symbol.name not in seen:
            sections.append(symbol.name)
            seen.add(symbol.name)
    order = {
        name: idx
        for idx, name in enumerate(
            ("meta", "strings", "condition", "events", "match", "outcome", "options")
        )
    }
    sections.sort(key=lambda name: (order.get(name, 99), name))
    ctx.set_cached(cache_key, sections)
    return sections



"""Expression rendering helpers for CodeGenerator."""

from __future__ import annotations


def visit_binary_expression(generator, node) -> str:
    return f"{generator.visit(node.left)} {node.operator} {generator.visit(node.right)}"


def visit_unary_expression(generator, node) -> str:
    operand = generator.visit(node.operand)
    if node.operator == "not":
        return f"not {operand}"
    return f"{node.operator}{operand}"


def visit_parentheses_expression(generator, node) -> str:
    return f"({generator.visit(node.expression)})"


def visit_set_expression(generator, node) -> str:
    return f"({', '.join(generator.visit(elem) for elem in node.elements)})"


def visit_range_expression(generator, node) -> str:
    return f"{generator.visit(node.low)}..{generator.visit(node.high)}"


def visit_function_call(generator, node) -> str:
    return f"{node.function}({', '.join(generator.visit(arg) for arg in node.arguments)})"


def visit_array_access(generator, node) -> str:
    return f"{generator.visit(node.array)}[{generator.visit(node.index)}]"


def visit_member_access(generator, node) -> str:
    return f"{generator.visit(node.object)}.{node.member}"


def visit_for_expression(generator, node) -> str:
    iterable = generator.visit(node.iterable)
    body = generator.visit(node.body)
    return f"for {node.quantifier} {node.variable} in {iterable} : ({body})"


def visit_at_expression(generator, node) -> str:
    return f"{node.string_id} at {generator.visit(node.offset)}"



"""AST dump visitor for CLI."""

from yaraast.ast.base import YaraFile
from yaraast.ast.conditions import (
    AtExpression,
    Condition,
    ForExpression,
    ForOfExpression,
    InExpression,
    OfExpression,
)
from yaraast.ast.expressions import (
    ArrayAccess,
    BinaryExpression,
    BooleanLiteral,
    DoubleLiteral,
    Expression,
    FunctionCall,
    Identifier,
    IntegerLiteral,
    MemberAccess,
    ParenthesesExpression,
    RangeExpression,
    SetExpression,
    StringCount,
    StringIdentifier,
    StringLength,
    StringLiteral,
    StringOffset,
    StringWildcard,
    UnaryExpression,
)
from yaraast.ast.meta import Meta
from yaraast.ast.modifiers import StringModifier
from yaraast.ast.rules import Import, Include, Rule, Tag
from yaraast.ast.strings import (
    HexAlternative,
    HexByte,
    HexJump,
    HexString,
    HexToken,
    HexWildcard,
    PlainString,
    RegexString,
    StringDefinition,
)
from yaraast.visitor import ASTVisitor


class ASTDumper(ASTVisitor[dict]):
    """Dump AST to dictionary format."""

    def visit_yara_file(self, node: YaraFile) -> dict:
        return {
            "type": "YaraFile",
            "imports": [self.visit(imp) for imp in node.imports],
            "includes": [self.visit(inc) for inc in node.includes],
            "rules": [self.visit(rule) for rule in node.rules],
        }

    def visit_import(self, node: Import) -> dict:
        return {"type": "Import", "module": node.module}

    def visit_include(self, node: Include) -> dict:
        return {"type": "Include", "path": node.path}

    def visit_rule(self, node: Rule) -> dict:
        tags = self._process_tags(node.tags)
        meta = self._process_meta(node.meta)
        modifiers = self._process_modifiers(node)

        return {
            "type": "Rule",
            "name": node.name,
            "modifiers": modifiers,
            "tags": tags,
            "meta": meta,
            "strings": [self.visit(s) for s in node.strings],
            "condition": self.visit(node.condition) if node.condition else None,
        }

    def _process_tags(self, tags) -> list:
        """Process tags from rule node."""
        result = []
        for tag in tags:
            if isinstance(tag, str):
                result.append(tag)
            else:
                result.append(self.visit(tag))
        return result

    def _process_meta(self, meta) -> dict:
        """Process meta from rule node."""
        result = {}
        if isinstance(meta, list):
            for m in meta:
                if hasattr(m, "key") and hasattr(m, "value"):
                    result[m.key] = m.value
        return result

    def _process_modifiers(self, node: Rule) -> list:
        """Process modifiers from rule node."""
        modifiers = []
        if hasattr(node, "modifiers") and node.modifiers:
            for mod in node.modifiers:
                if hasattr(mod, "accept"):
                    modifiers.append(self.visit(mod))
                else:
                    modifiers.append(str(mod))
        return modifiers

    def visit_tag(self, node: Tag) -> dict:
        return {"type": "Tag", "name": node.name}

    def visit_string_definition(self, node: StringDefinition) -> dict:
        return {"type": "StringDefinition", "identifier": node.identifier}

    def visit_plain_string(self, node: PlainString) -> dict:
        modifiers = self._extract_modifiers(node)
        return {
            "type": "PlainString",
            "identifier": node.identifier,
            "value": node.value,
            "modifiers": modifiers,
        }

    def visit_hex_string(self, node: HexString) -> dict:
        modifiers = self._extract_modifiers(node)
        return {
            "type": "HexString",
            "identifier": node.identifier,
            "tokens": [self.visit(token) for token in node.tokens],
            "modifiers": modifiers,
        }

    def visit_regex_string(self, node: RegexString) -> dict:
        modifiers = self._extract_modifiers(node)
        return {
            "type": "RegexString",
            "identifier": node.identifier,
            "regex": node.regex,
            "modifiers": modifiers,
        }

    def _extract_modifiers(self, node) -> list:
        """Extract modifiers from a string node."""
        modifiers = []
        if hasattr(node, "modifiers") and node.modifiers:
            for mod in node.modifiers:
                if hasattr(mod, "accept"):
                    modifiers.append(self.visit(mod))
                else:
                    modifiers.append(str(mod))
        return modifiers

    def visit_string_modifier(self, node: StringModifier) -> dict:
        return {"type": "StringModifier", "name": node.name, "value": node.value}

    def visit_hex_token(self, node: HexToken) -> dict:
        return {"type": "HexToken"}

    def visit_hex_byte(self, node: HexByte) -> dict:
        return {"type": "HexByte", "value": node.value}

    def visit_hex_wildcard(self, node: HexWildcard) -> dict:
        return {"type": "HexWildcard"}

    def visit_hex_jump(self, node: HexJump) -> dict:
        return {"type": "HexJump", "min_jump": node.min_jump, "max_jump": node.max_jump}

    def visit_hex_alternative(self, node: HexAlternative) -> dict:
        return {
            "type": "HexAlternative",
            "alternatives": [[self.visit(token) for token in alt] for alt in node.alternatives],
        }

    def visit_expression(self, node: Expression) -> dict:
        return {"type": "Expression"}

    def visit_identifier(self, node: Identifier) -> dict:
        return {"type": "Identifier", "name": node.name}

    def visit_string_identifier(self, node: StringIdentifier) -> dict:
        return {"type": "StringIdentifier", "name": node.name}

    def visit_string_wildcard(self, node: StringWildcard) -> dict:
        return {"type": "StringWildcard", "pattern": node.pattern}

    def visit_string_count(self, node: StringCount) -> dict:
        return {"type": "StringCount", "string_id": node.string_id}

    def visit_string_offset(self, node: StringOffset) -> dict:
        return {
            "type": "StringOffset",
            "string_id": node.string_id,
            "index": self.visit(node.index) if node.index else None,
        }

    def visit_string_length(self, node: StringLength) -> dict:
        return {
            "type": "StringLength",
            "string_id": node.string_id,
            "index": self.visit(node.index) if node.index else None,
        }

    def visit_integer_literal(self, node: IntegerLiteral) -> dict:
        return {"type": "IntegerLiteral", "value": node.value}

    def visit_double_literal(self, node: DoubleLiteral) -> dict:
        return {"type": "DoubleLiteral", "value": node.value}

    def visit_string_literal(self, node: StringLiteral) -> dict:
        return {"type": "StringLiteral", "value": node.value}

    def visit_boolean_literal(self, node: BooleanLiteral) -> dict:
        return {"type": "BooleanLiteral", "value": node.value}

    def visit_binary_expression(self, node: BinaryExpression) -> dict:
        return {
            "type": "BinaryExpression",
            "left": self.visit(node.left),
            "operator": node.operator,
            "right": self.visit(node.right),
        }

    def visit_unary_expression(self, node: UnaryExpression) -> dict:
        return {
            "type": "UnaryExpression",
            "operator": node.operator,
            "operand": self.visit(node.operand),
        }

    def visit_parentheses_expression(self, node: ParenthesesExpression) -> dict:
        return {
            "type": "ParenthesesExpression",
            "expression": self.visit(node.expression),
        }

    def visit_set_expression(self, node: SetExpression) -> dict:
        return {
            "type": "SetExpression",
            "elements": [self.visit(elem) for elem in node.elements],
        }

    def visit_range_expression(self, node: RangeExpression) -> dict:
        return {
            "type": "RangeExpression",
            "low": self.visit(node.low),
            "high": self.visit(node.high),
        }

    def visit_function_call(self, node: FunctionCall) -> dict:
        return {
            "type": "FunctionCall",
            "function": node.function,
            "arguments": [self.visit(arg) for arg in node.arguments],
        }

    def visit_array_access(self, node: ArrayAccess) -> dict:
        return {
            "type": "ArrayAccess",
            "array": self.visit(node.array),
            "index": self.visit(node.index),
        }

    def visit_member_access(self, node: MemberAccess) -> dict:
        return {
            "type": "MemberAccess",
            "object": self.visit(node.object),
            "member": node.member,
        }

    def visit_condition(self, node: Condition) -> dict:
        return {"type": "Condition"}

    def visit_for_expression(self, node: ForExpression) -> dict:
        return {
            "type": "ForExpression",
            "quantifier": node.quantifier,
            "variable": node.variable,
            "iterable": self.visit(node.iterable),
            "body": self.visit(node.body),
        }

    def visit_for_of_expression(self, node: ForOfExpression) -> dict:
        return {
            "type": "ForOfExpression",
            "quantifier": (
                node.quantifier
                if isinstance(node.quantifier, str | int)
                else self.visit(node.quantifier)
            ),
            "string_set": self.visit(node.string_set),
            "condition": self.visit(node.condition) if node.condition else None,
        }

    def visit_at_expression(self, node: AtExpression) -> dict:
        return {
            "type": "AtExpression",
            "string_id": node.string_id,
            "offset": self.visit(node.offset),
        }

    def visit_in_expression(self, node: InExpression) -> dict:
        return {
            "type": "InExpression",
            "string_id": node.string_id,
            "range": self.visit(node.range),
        }

    def visit_of_expression(self, node: OfExpression) -> dict:
        return {
            "type": "OfExpression",
            "quantifier": (
                node.quantifier
                if isinstance(node.quantifier, str | int)
                else self.visit(node.quantifier)
            ),
            "string_set": self.visit(node.string_set),
        }

    def visit_meta(self, node: Meta) -> dict:
        return {"type": "Meta", "key": node.key, "value": node.value}

    def visit_comment(self, node) -> dict:
        return {"type": "Comment", "text": node.text}

    def visit_comment_group(self, node) -> dict:
        return {"type": "CommentGroup", "lines": node.lines}

    def visit_defined_expression(self, node) -> dict:
        return {"type": "DefinedExpression", "expression": self.visit(node.expression)}

    def visit_dictionary_access(self, node) -> dict:
        return {
            "type": "DictionaryAccess",
            "object": self.visit(node.object),
            "key": node.key,
        }

    def visit_extern_import(self, node) -> dict:
        module_path = getattr(node, "module", None)
        if module_path is None:
            module_path = getattr(node, "module_path", None)
        return {"type": "ExternImport", "module": module_path}

    def visit_extern_namespace(self, node) -> dict:
        return {"type": "ExternNamespace", "name": node.name}

    def visit_extern_rule(self, node) -> dict:
        return {"type": "ExternRule", "name": node.name}

    def visit_extern_rule_reference(self, node) -> dict:
        rule_name = getattr(node, "name", None)
        if rule_name is None:
            rule_name = getattr(node, "rule_name", None)
        return {"type": "ExternRuleReference", "name": rule_name}

    def visit_hex_nibble(self, node) -> dict:
        return {"type": "HexNibble", "high": node.high, "value": node.value}

    def visit_in_rule_pragma(self, node) -> dict:
        return {"type": "InRulePragma", "directive": node.directive}

    def visit_module_reference(self, node) -> dict:
        return {"type": "ModuleReference", "module": node.module}

    def visit_pragma(self, node) -> dict:
        return {"type": "Pragma", "directive": node.directive}

    def visit_pragma_block(self, node) -> dict:
        return {"type": "PragmaBlock", "pragmas": [self.visit(p) for p in node.pragmas]}

    def visit_regex_literal(self, node) -> dict:
        return {
            "type": "RegexLiteral",
            "pattern": node.pattern,
            "modifiers": node.modifiers,
        }

    def visit_string_operator_expression(self, node) -> dict:
        return {
            "type": "StringOperatorExpression",
            "left": self.visit(node.left),
            "operator": node.operator,
            "right": self.visit(node.right),
        }



"""Outcome parsing for Enhanced YARA-L parser."""

from __future__ import annotations

from typing import Any

from yaraast.lexer.tokens import TokenType as BaseTokenType
from yaraast.yaral.ast_nodes import (
    AggregationFunction,
    ConditionalExpression,
    OutcomeAssignment,
    OutcomeExpression,
    OutcomeSection,
)
from yaraast.yaral.tokens import YaraLTokenType


class EnhancedYaraLParserOutcomeMixin:
    """Mixin for outcome parsing."""

    def _parse_outcome_section(self) -> OutcomeSection:
        """Parse enhanced outcome section with conditionals."""
        self._consume_keyword("outcome")
        self._consume(BaseTokenType.COLON, "Expected ':' after 'outcome'")

        assignments: list[OutcomeAssignment] = []

        while not self._check_section_keyword() and not self._check(BaseTokenType.RBRACE):
            if self._check_keyword("if"):
                cond_expr = self._parse_conditional_expression()
                assignments.append(
                    OutcomeAssignment(
                        variable="_",
                        expression=cond_expr,
                    ),
                )
            elif self._check_yaral_type(YaraLTokenType.EVENT_VAR) or self._check(
                BaseTokenType.STRING_IDENTIFIER,
            ):
                var_name = self._advance().value
                self._consume(BaseTokenType.EQ, "Expected '=' after outcome variable")
                expression = self._parse_outcome_expression()
                assignments.append(
                    OutcomeAssignment(
                        variable=var_name,
                        expression=expression,
                    ),
                )
            else:
                self._advance()

        return OutcomeSection(assignments=assignments)

    def _parse_outcome_expression(self) -> OutcomeExpression | Any:
        """Parse enhanced outcome expression with aggregations."""
        if self._check(BaseTokenType.IDENTIFIER):
            token = self._peek()
            if token.value in [
                "count",
                "count_distinct",
                "sum",
                "min",
                "max",
                "avg",
                "array",
                "array_distinct",
                "string_concat",
            ]:
                return self._parse_aggregation_function()

        if (
            self._check(BaseTokenType.IDENTIFIER)
            and self._peek_ahead(1)
            and self._peek_ahead(1).type == BaseTokenType.DOT
        ):
            return self._parse_udm_field_access()

        if self._check(BaseTokenType.STRING):
            return self._advance().value
        if self._check(BaseTokenType.INTEGER):
            return int(self._advance().value)

        raise self._error("Expected outcome expression")

    def _parse_aggregation_function(self) -> AggregationFunction:
        """Parse aggregation function."""
        func_name = self._advance().value

        self._consume(BaseTokenType.LPAREN, f"Expected '(' after {func_name}")

        arguments = []
        while not self._check(BaseTokenType.RPAREN):
            if self._check(BaseTokenType.IDENTIFIER):
                arg = self._parse_udm_field_access()
                arguments.append(arg)
            elif self._check(BaseTokenType.STRING) or self._check(BaseTokenType.INTEGER):
                arg = self._advance().value
                arguments.append(arg)

            if self._check(BaseTokenType.COMMA):
                self._advance()
            else:
                break

        self._consume(BaseTokenType.RPAREN, f"Expected ')' after {func_name} arguments")

        return AggregationFunction(function=func_name, arguments=arguments)

    def _parse_conditional_expression(self) -> ConditionalExpression:
        """Parse conditional expression in outcome section."""
        self._consume_keyword("if")

        condition = self._parse_condition_expression()

        self._consume_keyword("then")

        then_expr = self._parse_outcome_expression()

        else_expr = None
        if self._check_keyword("else"):
            self._advance()
            else_expr = self._parse_outcome_expression()

        return ConditionalExpression(
            condition=condition,
            true_value=then_expr,
            false_value=else_expr,
        )



"""Text search helpers for document highlight provider."""

from __future__ import annotations

from lsprotocol.types import DocumentHighlight, DocumentHighlightKind, Position, Range

from yaraast.lsp.structure import find_section_range, get_rule_text_range


def _is_identifier_boundary(line: str, start: int, end: int) -> bool:
    if start > 0 and (line[start - 1].isalnum() or line[start - 1] == "_"):
        return False
    return not (end < len(line) and (line[end].isalnum() or line[end] == "_"))


def simple_highlight(text: str, word: str) -> list[DocumentHighlight]:
    highlights = []
    lines = text.split("\n")
    for line_num, line in enumerate(lines):
        col = 0
        while True:
            idx = line.find(word, col)
            if idx == -1:
                break
            highlights.append(
                DocumentHighlight(
                    range=Range(
                        start=Position(line=line_num, character=idx),
                        end=Position(line=line_num, character=idx + len(word)),
                    ),
                    kind=DocumentHighlightKind.Text,
                )
            )
            col = idx + len(word)
    return highlights


def highlight_identifier(text: str, identifier: str) -> list[DocumentHighlight]:
    highlights = []
    lines = text.split("\n")
    for line_num, line in enumerate(lines):
        col = 0
        while True:
            idx = line.find(identifier, col)
            if idx == -1:
                break
            end_idx = idx + len(identifier)
            if not _is_identifier_boundary(line, idx, end_idx):
                col = idx + 1
                continue
            highlights.append(
                DocumentHighlight(
                    range=Range(
                        start=Position(line=line_num, character=idx),
                        end=Position(line=line_num, character=end_idx),
                    ),
                    kind=DocumentHighlightKind.Text,
                )
            )
            col = end_idx
    return highlights


def highlight_string_identifier(text: str, identifier: str) -> list[DocumentHighlight]:
    highlights = []
    lines = text.split("\n")
    base_id = identifier[1:] if identifier.startswith("$") else identifier
    patterns = [f"${base_id}", f"#{base_id}", f"@{base_id}", f"!{base_id}"]

    for line_num, line in enumerate(lines):
        for pattern in patterns:
            col = 0
            while True:
                idx = line.find(pattern, col)
                if idx == -1:
                    break
                end_idx = idx + len(pattern)
                if not _is_identifier_boundary(line, idx, end_idx):
                    col = end_idx
                    continue
                kind = DocumentHighlightKind.Read
                if pattern.startswith("$") and _is_string_definition_occurrence(
                    text, line_num, end_idx
                ):
                    kind = DocumentHighlightKind.Write
                highlights.append(
                    DocumentHighlight(
                        range=Range(
                            start=Position(line=line_num, character=idx),
                            end=Position(line=line_num, character=end_idx),
                        ),
                        kind=kind,
                    )
                )
                col = end_idx
    return highlights


def _is_string_definition_occurrence(text: str, line_num: int, end_idx: int) -> bool:
    rule_text_range = get_rule_text_range(text, line_num)
    if rule_text_range is None:
        return False
    section_range = find_section_range(
        rule_text_range.lines,
        "strings",
        rule_text_range.start,
        rule_text_range.end,
    )
    if section_range is None or not (
        section_range.start.line <= line_num <= section_range.end.line
    ):
        return False
    line = rule_text_range.lines[line_num]
    return line[end_idx:].lstrip().startswith("=")



"""Tests for YARA-L CLI formatting helpers."""

from __future__ import annotations

from yaraast.cli.yaral_services import _format_line
from yaraast.cli.yaral_services import format_yaral_code as _format_yaral_code


def test_format_line_handles_sections_and_braces() -> None:
    section_keywords = ["rule", "meta", "events", "match", "condition", "outcome", "options"]

    formatted, indent = _format_line("rule demo {", 0, section_keywords)
    assert formatted.startswith("rule demo {")
    assert indent == 1

    formatted, indent = _format_line("meta:", indent, section_keywords)
    assert formatted.strip() == "meta:"
    assert indent >= 1

    formatted, indent = _format_line("}", indent, section_keywords)
    assert formatted.strip() == "}"
    assert indent >= 0


def test_format_yaral_code_indents_sections() -> None:
    code = 'rule demo {\nmeta:\nauthor = "me"\ncondition:\n#e > 0\n}'
    formatted = _format_yaral_code(code)

    assert "rule demo {" in formatted
    assert "  meta:" in formatted
    assert '    author = "me"' in formatted
    assert "  condition:" in formatted



"""Additional branch coverage for hover provider internals (no mocks)."""

from __future__ import annotations

from types import SimpleNamespace

from lsprotocol.types import Position, Range

from yaraast.lsp.hover import HoverProvider


def _pos(line: int, char: int) -> Position:
    return Position(line=line, character=char)


def _range() -> Range:
    return Range(start=_pos(0, 0), end=_pos(0, 4))


def test_get_hover_none_and_dotted_non_member_paths() -> None:
    provider = HoverProvider()
    assert provider.get_hover("   ", _pos(0, 0)) is None
    assert provider.get_hover("a.b.c", _pos(0, 1)) is None
    hover = provider.get_hover("pe.imphash", _pos(0, 1))
    assert hover is not None
    assert "(function)" in hover.contents.value


def test_module_member_function_and_field_with_descriptions() -> None:
    provider = HoverProvider()
    provider.module_loader = SimpleNamespace(
        get_module=lambda name: (
            SimpleNamespace(
                functions={
                    "f": SimpleNamespace(
                        parameters=[("x", "int")],
                        return_type="bool",
                        description="does f",
                    ),
                },
                fields={
                    "k": SimpleNamespace(type="integer", description="field k"),
                },
            )
            if name == "m"
            else None
        ),
    )

    hover_f = provider._get_module_member_hover("m", "f", _range())
    assert hover_f is not None
    assert "function" in hover_f.contents.value
    assert "does f" in hover_f.contents.value

    hover_k = provider._get_module_member_hover("m", "k", _range())
    assert hover_k is not None
    assert "field" in hover_k.contents.value
    assert "field k" in hover_k.contents.value

    assert provider._get_module_member_hover("m", "missing", _range()) is None
    assert provider._get_module_member_hover("missing", "f", _range()) is None


def test_string_identifier_hover_regex_hex_and_fallback() -> None:
    provider = HoverProvider()

    text_regex = """
rule r {
  strings:
    $r = /ab+c/i
  condition:
    $r
}
""".lstrip()
    hover_regex = provider.get_hover(text_regex, _pos(4, 5))
    assert hover_regex is not None
    assert "(regex)" in hover_regex.contents.value

    text_hex = """
rule r {
  strings:
    $h = { 6A 40 ?? }
  condition:
    $h
}
""".lstrip()
    hover_hex = provider.get_hover(text_hex, _pos(4, 5))
    assert hover_hex is not None
    assert "(hex string)" in hover_hex.contents.value

    # Parse failure path falls back to generic identifier documentation.
    fallback = provider._get_string_identifier_hover("rule {", "$x", _range())
    assert fallback is not None
    assert "string identifier" in fallback.contents.value


def test_rule_hover_with_modifiers_tags_meta_and_strings() -> None:
    provider = HoverProvider()
    text = """
private rule alpha : t1 t2 {
  meta:
    author = "me"
  strings:
    $a = "abc"
  condition:
    true
}

rule beta {
  condition:
    alpha
}
""".lstrip()
    hover = provider.get_hover(text, _pos(11, 5))
    assert hover is not None
    value = hover.contents.value
    assert "(rule)" in value
    assert "private" in value
    assert "Tags: t1, t2" in value
    assert "Metadata" in value
    assert "**Strings:** 1 defined" in value

    # Parse failure path in _get_rule_hover should return None.
    assert provider._get_rule_hover("rule {", "x", _range()) is None


def test_get_hover_module_member_and_unknown_string_type_paths() -> None:
    provider = HoverProvider()

    provider.module_loader = SimpleNamespace(
        get_module=lambda _m: SimpleNamespace(
            functions={"f": SimpleNamespace(parameters=[], return_type="bool")},
            fields={},
        ),
    )
    hover = provider._get_module_member_hover("m", "f", _range())
    assert hover is not None
    assert "function" in hover.contents.value

    fallback = provider._get_string_identifier_hover("rule {", "$u", _range())
    assert fallback is not None
    assert "string identifier" in fallback.contents.value
